use std::collections::HashMap;
use std::fs;
use std::fs::File;

use bevy::math::Vec3;
use bevy::prelude::*;
use iyes_loopless::state::NextState;
use serde::{Deserialize, Serialize};

use crate::persistance::fs_utils::{get_root_save_dir, AUTOSAVE_NAME};
use crate::states::Player;
use crate::system::flow::AppState;
use crate::world::{
	ActiveLevel, CraftSpecialism, CurrentResidence, DistanceTravelled, EncounterState, HungerState,
	PendingLoadState, TownName, TradeManifestTickState, TradingState, TravelPath, TravelTarget,
};

pub type TownPersistenceData = HashMap<
	String,
	(
		TradingState,
		HungerState,
		TradeManifestTickState,
		Option<CraftSpecialism>,
	),
>;

#[derive(Serialize, Deserialize, Debug, Resource)]
pub struct PersistenceState {
	pub map_name: String,
	pub player_location: Vec3,
	pub player_inventory: TradingState,
	pub player_hunger: HungerState,
	#[serde(default)]
	pub travel_path: Option<TravelPath>,
	#[serde(default)]
	pub travel_target: Option<TravelTarget>,
	#[serde(default)]
	pub player_total_distance: DistanceTravelled,
	pub previous_location: CurrentResidence,
	pub encounter_state: EncounterState,
	pub town_states: TownPersistenceData,
}

#[derive(Clone, Debug)]
pub struct SaveFileEvent {
	pub filename: Option<String>,
}

impl SaveFileEvent {
	pub fn autosave() -> Self {
		Self { filename: None }
	}
}

#[derive(Clone, Debug)]
pub struct LoadFileEvent {
	pub filename: Option<String>,
}

impl LoadFileEvent {
	pub fn autosave() -> Self {
		Self { filename: None }
	}
}

type PlayerInfoQuery<'a, 'b> = Query<
	'a,
	'b,
	(
		&'static Transform,
		Option<&'static TravelPath>,
		Option<&'static TravelTarget>,
		&'static CurrentResidence,
		&'static TradingState,
		&'static HungerState,
		&'static DistanceTravelled,
	),
	With<Player>,
>;

pub fn sync_state_to_persistence(
	mut commands: Commands,
	mut state: Option<ResMut<PersistenceState>>,
	player_query: PlayerInfoQuery,
	town_query: Query<(
		&TownName,
		&TradingState,
		&HungerState,
		&TradeManifestTickState,
		Option<&CraftSpecialism>,
	)>,
	encounters: Res<EncounterState>,
	active_map: Res<ActiveLevel>,
) {
	let town_states = town_query
		.iter()
		.map(|(name, trading, hunger, manifest, maybe_spec)| {
			(
				name.0.clone(),
				(
					trading.clone(),
					hunger.clone(),
					manifest.clone(),
					maybe_spec.cloned(),
				),
			)
		})
		.collect::<TownPersistenceData>();

	match state {
		Some(mut state) => {
			if let Some((
				transform,
				maybe_travel,
				maybe_target,
				residence,
				trading,
				hunger,
				distance,
			)) = player_query.iter().next()
			{
				*state = PersistenceState {
					map_name: active_map.map.clone(),
					player_location: transform.translation,
					player_inventory: trading.clone(),
					player_hunger: hunger.clone(),
					player_total_distance: *distance,
					travel_path: maybe_travel.cloned(),
					travel_target: maybe_target.cloned(),
					previous_location: residence.clone(),
					encounter_state: encounters.clone(),
					town_states,
				};
			}
		}
		None => {
			if let Some((
				transform,
				maybe_travel,
				maybe_target,
				residence,
				trading,
				hunger,
				distance,
			)) = player_query.iter().next()
			{
				commands.insert_resource(PersistenceState {
					map_name: active_map.map.clone(),
					player_location: transform.translation,
					player_inventory: trading.clone(),
					player_hunger: hunger.clone(),
					player_total_distance: *distance,
					travel_path: maybe_travel.cloned(),
					travel_target: maybe_target.cloned(),
					previous_location: residence.clone(),
					encounter_state: encounters.clone(),
					town_states,
				})
			}
		}
	}
}

pub fn handle_save_event(
	mut events: ResMut<Events<SaveFileEvent>>,
	state: Option<Res<PersistenceState>>,
) {
	let root_data_dir = get_root_save_dir().expect("Could not find root dir for saves");

	if let Some(state) = state {
		for event in events.drain() {
			std::fs::create_dir_all(&root_data_dir).expect("Failed to create data dir");

			match fs::File::create(match event.filename {
				Some(name) => root_data_dir.join(name),
				None => root_data_dir.join(AUTOSAVE_NAME),
			}) {
				Ok(file) => {
					serde_json::to_writer_pretty(file, &*state)
						.inspect_err(|err| {
							log::error!("{}", err);
						})
						.expect("Failed to create save data");
				}
				Err(e) => {
					log::error!("{}", e);
				}
			}
		}
	}
}

pub fn handle_load_event(mut commands: Commands, mut events: ResMut<Events<LoadFileEvent>>) {
	let root_dir = get_root_save_dir().expect("Could not find root dir for saves");
	for event in events.drain() {
		let path = root_dir.join(event.filename.as_deref().unwrap_or(AUTOSAVE_NAME));
		match std::fs::File::open(path) {
			Ok(file) => match serde_json::from_reader::<File, PersistenceState>(file) {
				Ok(value) => {
					commands.insert_resource(PendingLoadState(value));
					commands.insert_resource(NextState(AppState::InGame));
				}
				Err(e) => {
					log::error!("Failed to parse save file: {}", e);
				}
			},
			Err(e) => {
				log::error!("Failed to load file: {}", e);
			}
		}
	}
}
