use std::collections::HashSet;
use std::hash::{Hash, Hasher};
use std::ops::Not;

use bevy::input::Input;
use bevy::prelude::{
	Color, Commands, Component, Entity, EventWriter, KeyCode, Query, Rect, Res, ResMut, Resource,
	Transform, With,
};
use bevy_prototype_lyon::draw::DrawMode;
use bevy_prototype_lyon::prelude::{FillMode, GeometryBuilder, RectangleOrigin, StrokeMode};
use bevy_prototype_lyon::shapes;
use iyes_loopless::state::NextState;

use crate::persistance::{LoadFileEvent, SaveFileEvent};
use crate::system::flow::AppState;
use crate::world::WorldZones;

#[derive(Debug, Copy, Clone, Resource, Ord, PartialOrd, Eq, PartialEq)]
pub enum DebugStatus {
	On,
	Off,
}

impl DebugStatus {
	pub fn inverse(&self) -> Self {
		match self {
			Self::Off => Self::On,
			Self::On => Self::Off,
		}
	}
}

impl Not for DebugStatus {
	type Output = DebugStatus;

	fn not(self) -> Self::Output {
		self.inverse()
	}
}

pub fn toggle_debug_mode(input: Res<Input<KeyCode>>, mut status: ResMut<DebugStatus>) {
	if input.just_released(KeyCode::F5) {
		*status = !*status;
	}
}

const DEBUG_SNAPSHOT: &str = "debug_snapshot.json";

pub fn handle_debug_snapshots(
	mut commands: Commands,
	input: Res<Input<KeyCode>>,
	mut save_events: EventWriter<SaveFileEvent>,
	mut load_events: EventWriter<LoadFileEvent>,
) {
	if input.just_released(KeyCode::F6) {
		save_events.send(SaveFileEvent {
			filename: Some(DEBUG_SNAPSHOT.to_string()),
		});
	} else if input.just_released(KeyCode::F7) {
		commands.insert_resource(NextState(AppState::Menu));
		load_events.send(LoadFileEvent {
			filename: Some(DEBUG_SNAPSHOT.to_string()),
		});
	}
}

#[derive(Component)]
pub struct ZoneShape;

#[derive(Component, Copy, Clone)]
pub struct ZoneData(Rect);
impl Eq for ZoneData {}
impl PartialEq for ZoneData {
	fn eq(&self, other: &Self) -> bool {
		other.0 == self.0
	}
}
impl Hash for ZoneData {
	fn hash<H: Hasher>(&self, state: &mut H) {
		state.write_i128((self.0.min.x * 1000.0) as i128);
		state.write_i128((self.0.min.y * 1000.0) as i128);
		state.write_i128((self.0.max.x * 1000.0) as i128);
		state.write_i128((self.0.max.y * 1000.0) as i128);
	}
}

pub fn show_zone_shapes(
	mut commands: Commands,
	query: Query<(Entity, &ZoneData), With<ZoneShape>>,
	debug: Res<DebugStatus>,
	zones: Res<WorldZones>,
) {
	if *debug == DebugStatus::Off {
		for (entity, _) in &query {
			commands.entity(entity).despawn();
		}
	} else {
		let mut handled = HashSet::with_capacity(zones.0.len());
		let map_zones = zones
			.0
			.iter()
			.map(|zone| ZoneData(zone.area))
			.collect::<HashSet<ZoneData>>();

		for (entity, zone) in &query {
			if map_zones.contains(zone) {
				handled.insert(*zone);
			} else {
				commands.entity(entity).despawn();
			}
		}

		for encounter in zones.0.iter() {
			let data = ZoneData(encounter.area);
			if !handled.contains(&data) {
				commands.spawn(GeometryBuilder::build_as(
					&shapes::Rectangle {
						origin: RectangleOrigin::Center,
						extents: data.0.half_size(),
					},
					DrawMode::Outlined {
						fill_mode: FillMode::color(Color::rgba(0.54, 0.23, 0.74, 0.25)),
						outline_mode: StrokeMode::color(Color::rgb(0.54, 0.23, 0.74)),
					},
					Transform::from_translation(data.0.center().extend(900.0)),
				));
			}
		}
	}
}
