use std::time::Instant;

use bevy::prelude::*;
use kayak_font::{KayakFont, TextProperties};
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	BackgroundBundle, ClipBundle, ElementBundle, NinePatch, TextProps, TextWidgetBundle,
};
use kayak_ui::DEFAULT_FONT;

use crate::assets::AssetHandles;
use crate::ui::prelude::{edge_px, px, stretch, value};

/// Props used by the [`TextBox`] widget
#[derive(Component, PartialEq, Debug, Clone)]
pub struct ATextBoxProps {
	/// If true, prevents the widget from being focused (and consequently edited)
	pub disabled: bool,
	/// The text to display when the user input is empty
	pub placeholder: Option<String>,
	/// The user input
	///
	/// This is a controlled state. You _must_ set this to the value to you wish to be displayed.
	/// You can use the [`on_change`] callback to update this prop as the user types.
	pub value: String,
	/// Set the size of the text to be rendered. This will determine element height
	pub font_size: f32,
}

impl Default for ATextBoxProps {
	fn default() -> Self {
		ATextBoxProps {
			disabled: Default::default(),
			placeholder: Default::default(),
			value: Default::default(),
			font_size: 18.0,
		}
	}
}

#[derive(Component, Clone, PartialEq)]
pub struct ATextBoxState {
	pub focused: bool,
	pub graphemes: Vec<String>,
	pub cursor_x: f32,
	pub cursor_position: usize,
	pub cursor_visible: bool,
	pub cursor_last_update: Instant,
	pub current_value: String,
}

impl Default for ATextBoxState {
	fn default() -> Self {
		Self {
			focused: Default::default(),
			graphemes: Default::default(),
			cursor_x: 0.0,
			cursor_position: Default::default(),
			cursor_visible: Default::default(),
			cursor_last_update: Instant::now(),
			current_value: String::new(),
		}
	}
}

pub struct ATextBoxValue(pub String);

impl Widget for ATextBoxProps {}

/// A widget that displays a text input field
/// A text box allows users to input text.
/// This text box is fairly simple and only supports basic input.
///
#[derive(Bundle)]
pub struct ATextBoxWidget {
	pub text_box: ATextBoxProps,
	pub styles: KStyle,
	pub computed_styles: ComputedStyles,
	pub on_event: OnEvent,
	pub on_layout: OnLayout,
	pub on_change: OnChange,
	pub focusable: Focusable,
	pub widget_name: WidgetName,
}

impl Default for ATextBoxWidget {
	fn default() -> Self {
		Self {
			text_box: Default::default(),
			styles: Default::default(),
			computed_styles: ComputedStyles::default(),
			on_event: Default::default(),
			on_layout: Default::default(),
			on_change: Default::default(),
			focusable: Default::default(),
			widget_name: ATextBoxProps::default().get_name(),
		}
	}
}

pub fn render_text_box_widget(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	mut query: Query<(
		&KStyle,
		&mut ComputedStyles,
		&ATextBoxProps,
		&mut OnEvent,
		&OnChange,
	)>,
	mut state_query: ParamSet<(Query<&ATextBoxState>, Query<&mut ATextBoxState>)>,
	font_assets: Res<Assets<KayakFont>>,
	font_mapping: Res<FontMapping>,
	asset_handles: Res<AssetHandles>,
) -> bool {
	if let Ok((styles, mut computed_styles, text_box, mut on_event, on_change)) =
		query.get_mut(entity)
	{
		let state_entity = widget_context.use_state::<ATextBoxState>(
			&mut commands,
			entity,
			ATextBoxState {
				current_value: text_box.value.clone(),
				..ATextBoxState::default()
			},
		);

		let mut is_different = false;
		if let Ok(state) = state_query.p0().get(state_entity) {
			if state.current_value != text_box.value {
				is_different = true;
			}
		}

		let style_font = styles.font.clone();
		let font_size = text_box.font_size;
		let inner_height = get_font_size_line_height(font_size);
		let element_height = inner_height + 24.0;

		if is_different {
			if let Ok(mut state) = state_query.p1().get_mut(state_entity) {
				state.current_value = text_box.value.clone();
				// Update graphemes
				set_graphemes(&mut state, &font_assets, &font_mapping, &style_font);

				state.cursor_position = state.graphemes.len();

				set_new_cursor_position(
					&mut state,
					&font_assets,
					&font_mapping,
					&style_font,
					font_size,
				);
			}
		}

		if let Ok(state) = state_query.p0().get(state_entity) {
			// let nine_vals = NinePatch {
			// 	handle: asset_handles.image("basic_panel"),
			// 	border: Edge::all(12.0),
			// };

			*computed_styles = KStyle::default()
				// Required styles
				.with_style(KStyle {
					render_command: value(match state.focused {
						true => RenderCommand::NinePatch {
							handle: asset_handles.image("panel_text_active"),
							border: Edge::all(12.0),
						},
						false => RenderCommand::NinePatch {
							handle: asset_handles.image("panel_text_idle"),
							border: Edge::all(12.0),
						}
					}),
					..Default::default()
				})
				// Apply any prop-given styles
				.with_style(styles)
				// If not set by props, apply these styles
				.with_style(KStyle {
					height: px(element_height),
					padding: edge_px(12.0),
					// cursor: CursorIcon::Text.into(),
					..Default::default()
				})
				.into();

			let background_styles = KStyle {
				render_command: StyleProp::Value(RenderCommand::Quad),
				height: px(get_font_size_line_height(font_size)),
				padding_left: px(5.0),
				padding_right: px(5.0),
				padding_bottom: px(4.0),
				..Default::default()
			};

			let cloned_on_change = on_change.clone();

			*on_event = OnEvent::new(
				move |In((event_dispatcher_context, _, mut event, _entity)): In<(
					EventDispatcherContext,
					WidgetState,
					Event,
					Entity,
				)>,
				      font_assets: Res<Assets<KayakFont>>,
				      font_mapping: Res<FontMapping>,
				      mut state_query: Query<&mut ATextBoxState>| {
					match event.event_type {
						EventType::KeyDown(key_event) => {
							if key_event.key() == KeyCode::Right {
								if let Ok(mut state) = state_query.get_mut(state_entity) {
									if state.cursor_position < state.graphemes.len() {
										state.cursor_position += 1;
									}
									set_new_cursor_position(
										&mut state,
										&font_assets,
										&font_mapping,
										&style_font,
										font_size,
									);
								}
							}
							if key_event.key() == KeyCode::Left {
								if let Ok(mut state) = state_query.get_mut(state_entity) {
									if state.cursor_position > 0 {
										state.cursor_position -= 1;
									}
									set_new_cursor_position(
										&mut state,
										&font_assets,
										&font_mapping,
										&style_font,
										font_size,
									);
								}
							}
						}
						EventType::CharInput { c } => {
							if let Ok(mut state) = state_query.get_mut(state_entity) {
								let cloned_on_change = cloned_on_change.clone();
								if !state.focused {
									return (event_dispatcher_context, event);
								}
								let cursor_pos = state.cursor_position;
								if is_backspace_key(c) {
									if !state.current_value.is_empty() {
										let char_pos: usize = state.graphemes[0..cursor_pos - 1]
											.iter()
											.map(|g| g.len())
											.sum();
										state.current_value.remove(char_pos);
										state.cursor_position -= 1;
									}
								} else if is_delete_key(c) {
									if state.cursor_position < state.graphemes.len() {
										let char_pos: usize = state.graphemes[0..cursor_pos]
											.iter()
											.map(|g| g.len())
											.sum();
										state.current_value.remove(char_pos);
										// state.cursor_position -= 1;
									}
								} else if !c.is_control() {
									let char_pos: usize = state.graphemes[0..cursor_pos]
										.iter()
										.map(|g| g.len())
										.sum();
									state.current_value.insert(char_pos, c);

									state.cursor_position += 1;
								}

								// Update graphemes
								set_graphemes(&mut state, &font_assets, &font_mapping, &style_font);

								set_new_cursor_position(
									&mut state,
									&font_assets,
									&font_mapping,
									&style_font,
									font_size,
								);
								cloned_on_change.set_value(state.current_value.clone());
								event.add_system(cloned_on_change);
							}
						}
						EventType::Focus => {
							if let Ok(mut state) = state_query.get_mut(state_entity) {
								state.focused = true;
								// Update graphemes
								set_graphemes(&mut state, &font_assets, &font_mapping, &style_font);

								state.cursor_position = state.graphemes.len();

								set_new_cursor_position(
									&mut state,
									&font_assets,
									&font_mapping,
									&style_font,
									font_size,
								);
							}
						}
						EventType::Blur => {
							if let Ok(mut state) = state_query.get_mut(state_entity) {
								state.focused = false;
							}
						}
						_ => {}
					}
					(event_dispatcher_context, event)
				},
			);

			let cursor_styles = KStyle {
				background_color: Color::ANTIQUE_WHITE.into(),
				position_type: value(KPositionType::SelfDirected),
				top: px(4.0),
				left: px(state.cursor_x),
				width: px(3.0),
				height: px(inner_height - 8.0),
				..Default::default()
			};

			let (text_value, text_color) = if !state.focused && text_box.value.is_empty() {
				(
					text_box.placeholder.clone().unwrap_or_default(),
					Color::GRAY,
				)
			} else {
				(state.current_value.clone(), Color::WHITE)
			};

			let text_styles = KStyle {
				color: value(text_color),
				top: stretch(1.0),
				bottom: px(4.0),
				..Default::default()
			};

			let shift = if let Some(layout) = widget_context.get_layout(entity) {
				let font_handle = match &styles.font {
					StyleProp::Value(font) => font_mapping.get_handle(font.clone()).unwrap(),
					_ => font_mapping.get_handle(DEFAULT_FONT.into()).unwrap(),
				};
				if let Some(font) = font_assets.get(&font_handle) {
					let string_to_cursor = state.graphemes[0..state.cursor_position].join("");
					let measurement = font.measure(
						&string_to_cursor,
						TextProperties {
							font_size,
							line_height: get_font_size_line_height(font_size),
							max_size: (10000.0, get_font_size_line_height(font_size) + 4.0),
							alignment: Alignment::Start,
							tab_size: 4,
						},
					);
					if measurement.size().0 > layout.width {
						(layout.width - measurement.size().0) - 20.0
					} else {
						0.0
					}
				} else {
					0.0
				}
			} else {
				0.0
			};

			let scroll_styles = KStyle {
				position_type: value(KPositionType::SelfDirected),
				padding_left: stretch(0.0),
				padding_right: stretch(0.0),
				padding_bottom: stretch(1.0),
				padding_top: stretch(1.0),
				left: px(shift),
				..Default::default()
			};

			let parent_id = Some(entity);
			rsx! {
				<BackgroundBundle styles={background_styles}>
					<ClipBundle styles={KStyle {
						height: px(get_font_size_line_height(font_size)),
						padding_left: stretch(0.0),
						padding_right: stretch(0.0),
						..Default::default()
					}}>
						<ElementBundle styles={scroll_styles}>
							<TextWidgetBundle
								styles={text_styles}
								text={TextProps {
									content: text_value,
									size: font_size,
									line_height: Some(get_font_size_line_height(font_size)),
									word_wrap: false,
									..Default::default()
								}}
							/>
							{
								if state.focused && state.cursor_visible {
									constructor! {
										<BackgroundBundle styles={cursor_styles} />
									}
								}
							}
						</ElementBundle>
					</ClipBundle>
				</BackgroundBundle>
			};
		}
	}

	true
}

/// Checks if the given character contains the "Backspace" sequence
///
/// Context: [Wikipedia](https://en.wikipedia.org/wiki/Backspace#Common_use)
fn is_backspace_key(c: char) -> bool {
	c == '\u{8}'
}

fn is_delete_key(c: char) -> bool {
	c == '\u{7f}'
}

fn set_graphemes(
	state: &mut ATextBoxState,
	font_assets: &Res<Assets<KayakFont>>,
	font_mapping: &FontMapping,
	style_font: &StyleProp<String>,
) {
	let font_handle = match style_font {
		StyleProp::Value(font) => font_mapping.get_handle(font.clone()).unwrap(),
		_ => font_mapping.get_handle(DEFAULT_FONT.into()).unwrap(),
	};

	if let Some(font) = font_assets.get(&font_handle) {
		state.graphemes = font
			.get_graphemes(&state.current_value)
			.iter()
			.map(|s| s.to_string())
			.collect::<Vec<_>>();
	}
}

fn get_single_grapheme_length(
	font_assets: &Res<Assets<KayakFont>>,
	font_mapping: &FontMapping,
	style_font: &StyleProp<String>,
	text: &String,
) -> usize {
	let font_handle = match style_font {
		StyleProp::Value(font) => font_mapping.get_handle(font.clone()).unwrap(),
		_ => font_mapping.get_handle(DEFAULT_FONT.into()).unwrap(),
	};

	if let Some(font) = font_assets.get(&font_handle) {
		let graphemes = font.get_graphemes(&text);
		return graphemes[0].len();
	}

	0
}

fn set_new_cursor_position(
	state: &mut ATextBoxState,
	font_assets: &Res<Assets<KayakFont>>,
	font_mapping: &FontMapping,
	style_font: &StyleProp<String>,
	font_size: f32,
) {
	let font_handle = match style_font {
		StyleProp::Value(font) => font_mapping.get_handle(font.clone()).unwrap(),
		_ => font_mapping.get_handle(DEFAULT_FONT.into()).unwrap(),
	};

	if let Some(font) = font_assets.get(&font_handle) {
		let string_to_cursor = state.graphemes[0..state.cursor_position].join("");
		let measurement = font.measure(
			&string_to_cursor,
			TextProperties {
				font_size,
				line_height: get_font_size_line_height(font_size),
				max_size: (10000.0, get_font_size_line_height(font_size) + 4.0),
				alignment: Alignment::Start,
				tab_size: 4,
			},
		);

		state.cursor_x = measurement.size().0;
	}
}

pub fn get_font_size_line_height(size: f32) -> f32 {
	if size <= 20.0 {
		size + 2.0
	} else {
		size + 4.0
	}
}

pub fn cursor_animation_system(
	mut state_query: ParamSet<(Query<(Entity, &ATextBoxState)>, Query<&mut ATextBoxState>)>,
) {
	let mut should_update = Vec::new();

	for (entity, state) in state_query.p0().iter() {
		if state.cursor_last_update.elapsed().as_secs_f32() > 0.5 && state.focused {
			should_update.push(entity);
		}
	}

	for state_entity in should_update.drain(..) {
		if let Ok(mut state) = state_query.p1().get_mut(state_entity) {
			state.cursor_last_update = Instant::now();
			state.cursor_visible = !state.cursor_visible;
		}
	}
}
