use bevy::prelude::*;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	KImage, KImageBundle, NinePatch, NinePatchBundle, TextureAtlasBundle, TextureAtlasProps,
};

use crate::assets::AssetHandles;
use crate::parent_widget;
use crate::ui::prelude::{px, stretch, value};

#[derive(Component, Clone, Eq, PartialEq, Default)]
pub enum ImageButtonContent {
	Image(String),
	Atlas(String, usize),
	#[default]
	None,
}

#[derive(Component, Clone, Eq, PartialEq, Default)]
pub struct ImageButtonWidgetProps {
	pub content: ImageButtonContent,
	pub is_disabled: bool,
	pub is_fixed: bool,
}

impl ImageButtonWidgetProps {
	pub fn image(name: impl ToString) -> Self {
		Self {
			content: ImageButtonContent::Image(name.to_string()),
			is_disabled: false,
			is_fixed: false,
		}
	}
	pub fn atlas(name: impl ToString, index: usize) -> Self {
		Self {
			content: ImageButtonContent::Atlas(name.to_string(), index),
			is_disabled: false,
			is_fixed: false,
		}
	}
}

impl Widget for ImageButtonWidgetProps {}

parent_widget!(ImageButtonWidgetProps => ImageButtonWidget);

#[derive(Component, PartialEq, Clone, Default)]
pub struct ImageButtonWidgetState {
	pub is_pressed: bool,
	pub is_hovered: bool,
}

pub struct ButtonClickEvent {
	pub button: Entity,
}

pub fn render_image_button_widget(
	In((mut widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	state_query: Query<&ImageButtonWidgetState>,
	mut query: Query<(
		&ImageButtonWidgetProps,
		&KChildren,
		&mut ComputedStyles,
		&KStyle,
	)>,
	assets: Res<AssetHandles>,
	atlass: Res<Assets<TextureAtlas>>,
) -> bool {
	if let Ok((props, children, mut computed, style)) = query.get_mut(entity) {
		let state_entity =
			widget_context.use_state(&mut commands, entity, ImageButtonWidgetState::default());

		let parent_id = Some(entity);

		if let Ok(state) = state_query.get(state_entity) {
			let events = OnEvent::new(
				move |In((event_dispatcher_context, _, mut event, _)): In<(
					EventDispatcherContext,
					WidgetState,
					Event,
					Entity,
				)>,
				      mut params: ParamSet<(
					Query<&ImageButtonWidgetProps>,
					Query<&mut ImageButtonWidgetState>,
				)>| {
					let widget_props = match params.p0().get(entity) {
						Ok(p) => p.clone(),
						Err(..) => return (event_dispatcher_context, event),
					};

					if let Ok(mut state) = params.p1().get_mut(state_entity) {
						match &event.event_type {
							EventType::Hover(..) | EventType::MouseIn(..) => {
								if !widget_props.is_disabled {
									state.is_hovered = true;
								}
							}
							EventType::MouseOut(..) => {
								state.is_hovered = false;
								state.is_pressed = false;
							}
							EventType::MouseDown(..) => {
								if !widget_props.is_disabled {
									state.is_pressed = true;
								}
							}
							EventType::MouseUp(..) => {
								state.is_pressed = false;
							}
							EventType::Click(..) => {
								if widget_props.is_disabled {
									event.prevent_default();
									event.stop_propagation();
								}
							}
							_ => {}
						}
					}

					(event_dispatcher_context, event)
				},
			);

			let nine_vals = if props.is_disabled {
				NinePatch {
					handle: assets.image("button_disabled"),
					border: Edge::all(3.0),
				}
			} else if state.is_pressed {
				NinePatch {
					handle: assets.image("button_down"),
					border: Edge::all(3.0),
				}
			} else if state.is_hovered {
				NinePatch {
					handle: assets.image("button_active"),
					border: Edge::all(3.0),
				}
			} else {
				NinePatch {
					handle: assets.image("button_idle"),
					border: Edge::all(3.0),
				}
			};

			let padding = if state.is_pressed {
				StyleProp::Value(Edge::new(
					Units::Pixels(12.0),
					Units::Pixels(8.0),
					Units::Pixels(12.0),
					Units::Pixels(8.0),
				))
			} else {
				StyleProp::Value(Edge::new(
					Units::Pixels(8.0),
					Units::Pixels(8.0),
					Units::Pixels(16.0),
					Units::Pixels(8.0),
				))
			};

			*computed = KStyle {
				render_command: value(RenderCommand::Layout),
				min_height: px(32.0),
				min_width: px(32.0),
				padding: value(Edge::all(Units::Stretch(0.0))),
				..Default::default()
			}
			.with_style(style)
			.into();

			let ninepatch_styles = KStyle {
				padding,
				..Default::default()
			};

			let image_styles = KStyle {
				background_color: Color::RED.into(),
				width: stretch(1.0),
				height: stretch(1.0),
				..Default::default()
			};

			rsx! {
				<NinePatchBundle
					on_event={events}
					nine_patch={nine_vals}
					styles={ninepatch_styles}
				>
					{match &props.content {
						ImageButtonContent::Image(name) => {
							let handle = assets.image(name);

							constructor! {
								<KImageBundle
									image={KImage(handle)}
									styles={image_styles}
								/>
							}
						},
						ImageButtonContent::Atlas(name, index) => {
							let atlas_handle = assets.atlas(name);
							let image_handle = assets.image(name);
							if let Some(atlas) = atlass.get(&atlas_handle) {
								let rect = atlas.textures[*index];
								let position = rect.min;
								let tile_size = rect.max - rect.min;

								constructor! {
									<TextureAtlasBundle
										atlas={TextureAtlasProps {
											tile_size,
											position,
											handle: image_handle,
										}}
										styles={image_styles}
									/>
								}

							}

						},
						ImageButtonContent::None => {}
					} }
				</NinePatchBundle>
			};
		}
	}
	true
}
