use bevy::prelude::*;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{KImage, KImageBundle, TextureAtlasBundle, TextureAtlasProps};
use serde::{Deserialize, Serialize};

use crate::assets::AssetHandles;
use crate::parent_widget;
use crate::ui::prelude::*;

#[derive(Clone, Default, Debug, Eq, PartialEq, Serialize, Deserialize)]
#[serde(untagged)]
pub enum IconContent {
	Image(String),
	Atlas(String, usize),
	#[default]
	None,
}

impl IconContent {
	pub fn is_none(&self) -> bool {
		match self {
			Self::None => true,
			_ => false,
		}
	}
}

#[derive(Clone, Component, PartialEq)]
pub struct InsetIconProps {
	pub image: IconContent,
	pub size: f32,
}

impl Default for InsetIconProps {
	fn default() -> Self {
		Self {
			image: IconContent::None,
			size: 48.0,
		}
	}
}

impl Widget for InsetIconProps {}

parent_widget!(InsetIconProps => InsetIconWidget);

pub fn render_inset_icon_widget(
	In((mut widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	mut query: Query<(&InsetIconProps, &mut ComputedStyles, &KStyle)>,
	assets: Res<AssetHandles>,
	atlass: Res<Assets<TextureAtlas>>,
) -> bool {
	let parent_id = Some(entity);

	if let Ok((props, mut computed, style)) = query.get_mut(entity) {
		*computed = KStyle {
			render_command: value(RenderCommand::Layout),
			min_height: px(32.0),
			min_width: px(32.0),
			height: px(props.size),
			width: px(props.size),
			padding: value(Edge::all(Units::Stretch(0.0))),
			..Default::default()
		}
		.with_style(style)
		.into();

		let image_styles = KStyle {
			width: stretch(1.0),
			height: stretch(1.0),
			..Default::default()
		};

		match &props.image {
			IconContent::Image(name) => {
				let handle = assets.image(name);

				rsx! {
					<KImageBundle
						image={KImage(handle)}
						styles={image_styles}
					/>
				};
			}
			IconContent::Atlas(name, index) => {
				let atlas_handle = assets.atlas(name);
				let image_handle = assets.image(name);
				if let Some(atlas) = atlass.get(&atlas_handle) {
					let rect = atlas.textures[*index];
					let position = rect.min;
					let tile_size = rect.max - rect.min;

					rsx! {
						<TextureAtlasBundle
							atlas={TextureAtlasProps {
								tile_size,
								position,
								handle: image_handle,
							}}
							styles={image_styles}
						/>
					};
				}
			}
			IconContent::None => {}
		}
	}

	true
}
