use bevy::prelude::*;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{BackgroundBundle, TextProps, TextWidgetBundle};

use crate::ui::components::*;
use crate::ui::prelude::{px, stretch, value};
use crate::ui::utilities::TextSizer;
use crate::ui::widgets::*;
use crate::{basic_widget, empty_props};

#[derive(Component, Debug, Clone, PartialEq)]
pub struct PanelStatProps {
	pub icon: IconContent,
	pub text: String,
	pub font_size: f32,
	pub color: Color,
	pub fill: bool,
}

impl Default for PanelStatProps {
	fn default() -> Self {
		Self {
			icon: IconContent::None,
			text: String::new(),
			font_size: 18.0,
			color: Color::BLACK,
			fill: false,
		}
	}
}

impl Widget for PanelStatProps {}

basic_widget!(PanelStatProps => PanelStatWidget);

pub fn render_panel_stat_widget(
	In((mut widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	mut query: Query<(&PanelStatProps, &mut ComputedStyles, &KStyle)>,
	text_sizer: TextSizer,
) -> bool {
	let parent_id = Some(entity);

	if let Ok((props, mut computed, style)) = query.get_mut(entity) {
		let (text_width, text_height) =
			text_sizer.measure_text(props.text.as_str(), props.font_size);

		let between = 10.0;
		let image_size = props.font_size;

		let total_width = text_width + between + image_size;

		*computed = if props.fill {
			KStyle {
				render_command: value(RenderCommand::Quad),
				min_width: px(total_width),
				width: stretch(1.0),
				height: px(text_height),
				..Default::default()
			}
			.with_style(style)
			.into()
		} else {
			KStyle {
				render_command: value(RenderCommand::Quad),
				width: px(total_width),
				height: px(text_height),
				..Default::default()
			}
			.with_style(style)
			.into()
		};

		let wrapper_style = KStyle {
			layout_type: value(LayoutType::Row),
			col_between: if props.fill {
				stretch(1.0)
			} else {
				px(between)
			},
			..Default::default()
		};

		rsx! {
			<BackgroundBundle styles={wrapper_style}>
				<InsetIconWidget props={InsetIconProps { image: props.icon.clone(), size: image_size }} />
				<TextWidgetBundle
					text={TextProps {
						content: props.text.clone(),
						size: props.font_size,
						line_height: Some(props.font_size + 2.0),
						..Default::default()
					}}
					styles={KStyle {
						color: value(props.color),
						..Default::default()
					}}
				/>
			</BackgroundBundle>
		};
	}

	true
}
