use bevy::app::{PluginGroup, PluginGroupBuilder};
use kayak_ui::prelude::KayakContextPlugin;
use kayak_ui::widgets::KayakWidgets;

pub mod components;
pub mod context;
pub mod screens;
pub mod sync;
pub mod utilities;
pub mod widgets;

pub mod prelude {
	use bevy::prelude::FromReflect;
	use kayak_ui::prelude::{Alignment, Edge, StyleProp, Units};

	#[inline(always)]
	pub fn px(val: f32) -> StyleProp<Units> {
		StyleProp::Value(Units::Pixels(val))
	}

	#[inline(always)]
	pub fn edge_px(val: f32) -> StyleProp<Edge<Units>> {
		StyleProp::Value(Edge::all(Units::Pixels(val)))
	}

	#[inline(always)]
	pub fn pct(val: f32) -> StyleProp<Units> {
		StyleProp::Value(Units::Percentage(val))
	}

	#[inline(always)]
	pub fn stretch(val: f32) -> StyleProp<Units> {
		StyleProp::Value(Units::Stretch(val))
	}

	#[inline(always)]
	pub fn val_auto() -> StyleProp<Units> {
		StyleProp::Value(Units::Auto)
	}

	#[inline(always)]
	pub fn value<T: Clone + Default + FromReflect>(val: T) -> StyleProp<T> {
		StyleProp::Value(val)
	}

	#[inline(always)]
	pub fn simple_text<Value: ToString, Font: ToString>(
		value: Value,
		font: Font,
		size: f32,
	) -> kayak_ui::widgets::TextProps {
		kayak_ui::widgets::TextProps {
			content: value.to_string(),
			font: Some(font.to_string()),
			show_cursor: true,
			size,
			alignment: Alignment::Start,
			..Default::default()
		}
	}
}

mod _config {
	use bevy::app::{App, Plugin};
	use bevy::prelude::{Commands, DespawnRecursiveExt, Entity, Local, Query, Res, ResMut, With};
	use iyes_loopless::prelude::{AppLooplessStateExt, ConditionSet};
	use kayak_ui::prelude::FontMapping;

	use crate::assets::AssetHandles;
	use crate::system::flow::AppState;
	use crate::ui::utilities::StateUIRoot;

	pub fn configure_kayak_ui(
		mut commands: Commands,
		assets: Res<AssetHandles>,
		mut font_mapping: ResMut<FontMapping>,
	) {
		log::info!("Configuring Fonts And UI");
		font_mapping.set_default(assets.kayak_font("equipment_pro"));
		font_mapping.add("header", assets.kayak_font("compass_pro"));
	}

	pub fn remove_ui(mut commands: Commands, query: Query<Entity, With<StateUIRoot>>) {
		for entity in &query {
			commands.entity(entity).despawn_recursive();
		}
	}

	pub struct ConfigureKayakPlugin;
	impl Plugin for ConfigureKayakPlugin {
		fn build(&self, app: &mut App) {
			app.add_exit_system(AppState::Setup, configure_kayak_ui)
				.add_enter_system(AppState::Menu, super::screens::render_menu_ui)
				.add_enter_system(AppState::InGame, super::screens::render_in_game_ui)
				.add_enter_system(AppState::Tips, super::screens::render_menu_tips_ui)
				.add_enter_system(AppState::Settings, super::screens::render_menu_settings_ui)
				.add_exit_system(AppState::InGame, remove_ui)
				.add_exit_system(AppState::Tips, remove_ui)
				.add_exit_system(AppState::Settings, remove_ui)
				.add_exit_system(AppState::Menu, remove_ui);
		}
	}
}

pub struct AdventUIPlugins;
impl PluginGroup for AdventUIPlugins {
	fn build(self) -> PluginGroupBuilder {
		PluginGroupBuilder::start::<Self>()
			// .add(kayak_ui::bevy::BevyKayakUIPlugin)
			.add(_config::ConfigureKayakPlugin)
			.add(KayakContextPlugin)
			.add(KayakWidgets)
			.add(sync::UISyncPlugin)
	}
}
