use bevy::prelude::*;
use iyes_loopless::prelude::NextState;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	ElementBundle, KImage, KImageBundle, KayakAppBundle, TextProps, TextWidgetBundle,
};

use crate::assets::AssetHandles;
use crate::persistance::{fs_utils, LoadFileEvent};
use crate::system::flow::AppState;
use crate::ui::components::*;
use crate::ui::context::*;
use crate::ui::prelude::{pct, px, stretch, value};
use crate::ui::sync::UITravelInfo;
use crate::ui::utilities::context::create_root_context;
use crate::ui::utilities::StateUIRoot;
use crate::ui::widgets::*;
use crate::world::EncounterState;
use crate::{
	empty_props, on_button_click, parent_widget, register_widget, register_widget_with_context,
	register_widget_with_many_resources, register_widget_with_resource,
};

empty_props!(MainMenuProps);
parent_widget!(MainMenuProps => MainMenuLayout);

#[derive(Component, PartialEq, Eq, Clone, Default)]
pub struct MainMenuState {
	bg_idx: usize,
}

pub fn render_main_menu_layout(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	assets: Res<AssetHandles>,
	state_query: Query<&MainMenuState>,
) -> bool {
	let parent_id = Some(entity);
	let has_autosave = fs_utils::has_auto_save();

	let state_entity = widget_context.use_state(
		&mut commands,
		entity,
		MainMenuState {
			bg_idx: fastrand::usize(1..6),
		},
	);

	if let Ok(state) = state_query.get(state_entity) {
		log::info!("Rendering main menu");

		let root_styles = KStyle {
			..Default::default()
		};

		let image_styles = KStyle {
			position_type: value(KPositionType::SelfDirected),
			..Default::default()
		};

		let contents_container_style = KStyle {
			position_type: value(KPositionType::SelfDirected),
			width: pct(65.0),
			height: pct(80.0),
			min_width: px(400.0),
			min_height: px(300.0),
			max_width: px(500.0),
			max_height: px(400.0),
			top: stretch(1.0),
			left: stretch(1.0),
			right: stretch(1.0),
			bottom: stretch(1.0),
			layout_type: value(LayoutType::Column),
			row_between: px(20.0),
			..Default::default()
		};

		let header_style = KStyle {
			left: stretch(1.0),
			right: stretch(1.0),
			..Default::default()
		};

		let image = KImage(assets.image(format!("menu_background_{}", state.bg_idx)));

		let button_style = KStyle {
			width: px(300.0),
			left: stretch(1.0),
			right: stretch(1.0),
			..Default::default()
		};

		rsx! {
			<ElementBundle styles={root_styles}>
				<KImageBundle styles={image_styles} image={image} />
				<MainMenuPanel />
			</ElementBundle>
		};
	}

	true
}

pub fn render_menu_ui(mut commands: Commands) {
	let parent_id = None;
	let mut widget_context = create_main_menu_context();

	rsx! {
		<KayakAppBundle>
			<MainMenuLayout />
		</KayakAppBundle>
	};

	commands.spawn((UICameraBundle::new(widget_context), StateUIRoot));
}

fn create_main_menu_context() -> KayakRootContext {
	let mut widget_context = create_root_context();

	register_widget_with_context!(
		widget_context,
		MainMenuProps,
		MainMenuState,
		MainMenuContext,
		render_main_menu_layout
	);

	register_widget_with_context!(
		widget_context,
		MainMenuContextProps,
		EmptyState,
		MainMenuContext,
		render_main_menu_context_provider
	);

	widget_context
}
