use std::collections::HashMap;

use bevy::asset::{Assets, Handle};
use bevy::prelude::{Entity, Query, Res, ResMut, Resource, With};

use crate::const_data::{get_goods_from_name, get_goods_from_name_checked, TRADE_GOODS};
use crate::states::Player;
use crate::world::{
	CurrentResidence, HungerState, StarvationMarker, TownName, TradeGood, TradeManifest,
	TradingState, TravelTarget,
};

#[derive(Clone)]
pub struct ItemTradeData {
	pub definition: TradeGood,
	pub current_cost: usize,
	pub town_amount: usize,
	pub player_amount: usize,
}

#[derive(Resource, Default)]
pub struct UITradeData {
	pub town_gold: usize,
	pub town_entity: Option<Entity>,
	pub player_gold: usize,
	pub player_entity: Option<Entity>,
	pub shop_items: Vec<ItemTradeData>,
	pub hunger_descriptor: String,
	pub is_starved: bool,
}

pub fn sync_ui_trade_data(
	player_query: Query<
		(
			Entity,
			&TradingState,
			Option<&TravelTarget>,
			&CurrentResidence,
		),
		With<Player>,
	>,
	town_query: Query<(
		Entity,
		&TownName,
		&TradingState,
		&HungerState,
		&Handle<TradeManifest>,
		Option<&StarvationMarker>,
	)>,
	manifests: Res<Assets<TradeManifest>>,
	mut trade_data: ResMut<UITradeData>,
) {
	let mut trade_entries = HashMap::with_capacity(15);

	let (entity, player_trading_state, target, residence) = match player_query.get_single() {
		Ok(p) => p,
		Err(_) => return,
	};

	trade_data.player_entity = Some(entity);
	trade_data.player_gold = player_trading_state.gold.max(0) as usize;

	let town_name = target
		.map(|target| target.0.clone())
		.unwrap_or_else(|| match residence {
			CurrentResidence::RestingAt(place) => place.clone(),
			CurrentResidence::TravellingFrom(..) => String::new(),
		});

	let (town_entity, _, town_trading_state, hunger_state, manifest_handle, starvation) =
		match town_query
			.into_iter()
			.find(|(_, name, _, _, _, _)| town_name == *name.0)
		{
			Some(v) => v,
			None => {
				trade_data.shop_items = vec![];
				trade_data.town_entity = None;
				return;
			}
		};

	trade_data.town_entity = Some(town_entity);
	trade_data.town_gold = town_trading_state.gold.max(0) as usize;
	trade_data.hunger_descriptor = hunger_state.get_town_descriptor();
	trade_data.is_starved = starvation.is_some();

	for (item_name, trade_good) in TRADE_GOODS.iter() {
		let manifest = match manifests.get(manifest_handle) {
			Some(m) => m,
			None => continue,
		};

		let roster_entry = match manifest.get(&trade_good.name) {
			Some(e) => e,
			None => continue,
		};

		let town_amount = match town_trading_state.items.get(&trade_good.name) {
			Some(q) => *q,
			None => 0,
		};

		let player_amount = match player_trading_state.items.get(&trade_good.name) {
			Some(q) => *q,
			None => 0,
		};

		let mut cost_multiplier = -1.0;

		for (limit, multiplier) in &roster_entry.cost_multipliers {
			cost_multiplier = *multiplier;
			if town_amount < *limit {
				break;
			}
		}

		let current_fractional_cost = cost_multiplier * trade_good.gold_value as f32;
		let current_cost = current_fractional_cost.ceil() as usize;

		trade_entries.insert(
			item_name.clone(),
			ItemTradeData {
				definition: trade_good.clone(),
				current_cost,
				town_amount,
				player_amount,
			},
		);
	}

	let mut entries = trade_entries
		.into_values()
		.filter(|listing| {
			listing.current_cost > 0 && (listing.town_amount > 0 || listing.player_amount > 0)
		})
		.collect::<Vec<ItemTradeData>>();
	entries.sort_by(|a, b| a.definition.name.cmp(&b.definition.name));
	trade_data.shop_items = entries;
}
