use bevy::prelude::*;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	BackgroundBundle, ElementBundle, ScrollBoxBundle, ScrollBoxProps, ScrollContextProviderBundle,
	TextProps, TextWidgetBundle,
};

use crate::ui::components::*;
use crate::ui::prelude::*;
use crate::ui::widgets::*;
use crate::world::{EncounterOption, EncounterOutcome, EncounterState, SelectEncounterOptionEvent};
use crate::{basic_widget, empty_props, on_button_click};

empty_props!(EncounterPanelProps);
basic_widget!(EncounterPanelProps => EncounterPanel);

pub fn render_encounter_panel(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	ui_data: Res<EncounterState>,
) -> bool {
	let parent_id = Some(entity);

	let distance_style = KStyle {
		position_type: value(KPositionType::SelfDirected),
		top: stretch(1.0),
		left: stretch(1.0),
		right: stretch(1.0),
		bottom: px(50.0),
		..Default::default()
	};

	let panel_style = KStyle {
		position_type: value(KPositionType::SelfDirected),
		width: pct(65.0),
		height: pct(80.0),
		min_width: px(400.0),
		min_height: px(300.0),
		max_width: px(600.0),
		max_height: px(550.0),
		top: stretch(1.0),
		left: stretch(1.0),
		right: stretch(1.0),
		bottom: stretch(1.0),
		layout_type: value(LayoutType::Column),
		row_between: px(20.0),
		..Default::default()
	};

	let encounter_data = match &*ui_data {
		EncounterState::NoEncounter => None,
		EncounterState::Choice(encounter) => Some((
			encounter.title.clone(),
			encounter.description.clone(),
			encounter.options.clone(),
		)),
		EncounterState::Consequence(encounter, choice) => Some((
			encounter.title.clone(),
			choice.describe(),
			vec![EncounterOption {
				label: String::from("Continue"),
				requirements: vec![],
				hide_missing_reqs: false,
				outcomes: vec![EncounterOutcome::Continue],
			}],
		)),
	};

	if let Some((title, description, options)) = encounter_data {
		log::info!("{}", description);

		rsx! {
			<ElementBundle styles={panel_style}>
				<PanelWidget>
					<BackgroundBundle
						styles={KStyle {

							height: value(Units::Auto),
							color: value(Color::BLACK),
							padding_left: px(20.0),
							padding_right: px(20.0),
							padding_top: px(10.0),
							padding_bottom: px(10.0),
							..Default::default()
						}}
					>
						<TextWidgetBundle
							text={TextProps {
								font: Some(String::from("header")),
								content: format!("Trepidation! {}", title),
								size: 36.0,
								..Default::default()
							}}
							styles={KStyle {
								color: value(Color::BLACK),
								left: stretch(1.0),
								right: stretch(1.0),
								..Default::default()
							}}
						/>
					</BackgroundBundle>
					<VDividerWidget props={VDividerWidgetProps { height: 4.0, padding: 10.0, color: Color::rgb(0.52, 0.369, 0.18)}} />
					<ScrollContextProviderBundle>
						<ScrollBoxBundle
							scroll_box_props={ScrollBoxProps {
								track_color: Some(Color::rgb(0.827, 0.482, 0.353)),
								thumb_color: Some(Color::rgb(0.451, 0.224, 0.063)),
								..Default::default()
							} } styles={KStyle { padding: edge_px(5.0), ..Default::default() }}>
								{ for line in description.split("\n") {
									constructor! {
										<BackgroundBundle
											styles={KStyle {

												pointer_events: value(PointerEvents::None),
												render_command: value(RenderCommand::Quad),
												layout_type: value(LayoutType::Row),
												height: value(Units::Auto),
												width: pct(90.0),
												left: stretch(1.0),
												right: stretch(1.0),
												bottom: px(15.0),
												..Default::default()
											}}
										>
											<TextWidgetBundle
												text={TextProps {
													content: line.to_string(),
													line_height: Some(34.0),
													size: 32.0,
													..Default::default()
												}}
												styles={KStyle {
													color: value(Color::BLACK),
													..Default::default()
												}}
											/>
										</BackgroundBundle>
									}
								} }
						</ScrollBoxBundle>
					</ScrollContextProviderBundle>
				</PanelWidget>

				<ElementBundle styles={KStyle {
					layout_type: value(LayoutType::Row),
					padding_top: stretch(1.0),
					padding_bottom: stretch(1.0),
					padding_left: stretch(1.0),
					padding_right: stretch(1.0),
					height: px(60.0),
					col_between: px(15.0),
					..Default::default()
				}}>
					{
						for option in &options {
							let label = format!("{}", option.label);
							let owned = option.clone();
							let on_click = on_button_click!(EventWriter<SelectEncounterOptionEvent>, |mut state: EventWriter<SelectEncounterOptionEvent>| {
								state.send(SelectEncounterOptionEvent { option: owned.clone() });
							});

							constructor! {
								<ButtonWidget
									styles={KStyle { ..Default::default() }}
									props={ButtonWidgetProps::text(label, 28.0)}
									on_event={on_click}
								/>
							}
						}
					}
				</ElementBundle>
			</ElementBundle>
		};
	}

	true
}
