use bevy::prelude::*;
use iyes_loopless::prelude::NextState;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	BackgroundBundle, ElementBundle, KImage, KImageBundle, TextProps, TextWidgetBundle,
};
use micro_musicbox::prelude::AudioSettings;

use crate::assets::AssetHandles;
use crate::persistance::fs_utils::{save_config_file, Options};
use crate::persistance::{fs_utils, LoadFileEvent};
use crate::system::flow::AppState;
use crate::ui::components::*;
use crate::ui::context::{MainMenuContext, MainMenuView};
use crate::ui::prelude::{edge_px, pct, px, stretch, value};
use crate::ui::widgets::*;
use crate::{basic_widget, empty_props, on_button_click};

empty_props!(SettingsPanelProps);
basic_widget!(SettingsPanelProps => SettingsPanel);

#[derive(Default, Component, Clone, Copy, Eq, PartialEq)]
pub struct SettingsPanelState {
	tip_index: usize,
}

pub fn render_settings_panel(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	assets: Res<AssetHandles>,
	audio: Res<AudioSettings>,
) -> bool {
	let parent_id = Some(entity);

	let state_entity =
		widget_context.use_state(&mut commands, entity, SettingsPanelState::default());

	let contents_container_style = KStyle {
		position_type: value(KPositionType::SelfDirected),
		width: pct(65.0),
		height: pct(80.0),
		min_width: px(400.0),
		min_height: px(300.0),
		max_width: px(500.0),
		max_height: px(400.0),
		top: stretch(1.0),
		left: stretch(1.0),
		right: stretch(1.0),
		bottom: stretch(1.0),
		layout_type: value(LayoutType::Column),
		row_between: px(20.0),
		color: value(Color::BLACK),
		..Default::default()
	};

	let header_style = KStyle {
		left: stretch(1.0),
		right: stretch(1.0),
		bottom: px(30.0),
		..Default::default()
	};

	let tip_style = KStyle {
		height: stretch(1.0),
		padding_left: px(20.0),
		padding_right: px(20.0),
		..Default::default()
	};

	let button_style = KStyle {
		..Default::default()
	};

	let controls_style = KStyle {
		top: stretch(1.0),
		layout_type: value(LayoutType::Row),
		col_between: stretch(1.0),
		..Default::default()
	};

	let row_style = KStyle {
		layout_type: value(LayoutType::Row),
		col_between: px(10.0),
		height: px(60.0),
		padding_top: stretch(1.0),
		padding_bottom: stretch(1.0),
		..Default::default()
	};

	rsx! {
		<ElementBundle  styles={contents_container_style}>
			<PanelWidget>
				<TextWidgetBundle
					text={TextProps {
						content: String::from("Settings"),
						font: Some(String::from("header")),
						size: 52.0,
						..Default::default()
					}}
					styles={header_style}
				/>

				<BackgroundBundle
					styles={row_style.clone()}
				>
					<TextWidgetBundle
						text={TextProps {
							content: String::from("Master Volume"),
							size: 32.0,
							..Default::default()
						}}
					/>
					<BackgroundBundle styles={KStyle { width: stretch(1.0), ..Default::default() }} />
					<ButtonWidget
						props={ButtonWidgetProps {
							text: String::from("-"),
							font_size: 24.0,
							..Default::default()
						}}
						on_event={on_button_click!(ResMut<AudioSettings>, |mut settings: ResMut<
							AudioSettings,
						>| {
							settings.master_volume = (settings.master_volume - 0.05).max(0.0);
							save_config_file(Options {
								audio: *settings
							});
						})}
					/>
					<BackgroundBundle
						styles={KStyle {
							padding_left: stretch(1.0),
							padding_right: stretch(1.0),
							padding_top: stretch(1.0),
							padding_bottom: stretch(1.0),
							width: px(100.0),
							..Default::default()
						}}
					>
						<TextWidgetBundle
							text={TextProps {
								content: format!("{:.0}", audio.master_volume * 100.0),
								size: 32.0,
								..Default::default()
							}}
						/>
					</BackgroundBundle>
					<ButtonWidget
						props={ButtonWidgetProps {
							text: String::from("+"),
							font_size: 24.0,
							..Default::default()
						}}
						on_event={on_button_click!(ResMut<AudioSettings>, |mut settings: ResMut<
							AudioSettings,
						>| {
							settings.master_volume = (settings.master_volume + 0.05).min(1.0);
							save_config_file(Options {
								audio: *settings
							});
						})}
					/>
				</BackgroundBundle>
				<BackgroundBundle
					styles={row_style.clone()}
				>
					<TextWidgetBundle
						text={TextProps {
							content: String::from("Music Volume"),
							size: 32.0,
							..Default::default()
						}}
					/>
					<BackgroundBundle styles={KStyle { width: stretch(1.0), ..Default::default() }} />
					<ButtonWidget
						props={ButtonWidgetProps {
							text: String::from("-"),
							font_size: 24.0,
							..Default::default()
						}}
						on_event={on_button_click!(ResMut<AudioSettings>, |mut settings: ResMut<
							AudioSettings,
						>| {
							settings.music_volume = (settings.music_volume - 0.05).max(0.0);
							save_config_file(Options {
								audio: *settings
							});
						})}
					/>
					<BackgroundBundle
						styles={KStyle {
							padding_left: stretch(1.0),
							padding_right: stretch(1.0),
							padding_top: stretch(1.0),
							padding_bottom: stretch(1.0),
							width: px(100.0),
							..Default::default()
						}}
					>
						<TextWidgetBundle
							text={TextProps {
								content: format!("{:.0}", audio.music_volume * 100.0),
								size: 32.0,
								..Default::default()
							}}
						/>
					</BackgroundBundle>
					<ButtonWidget
						props={ButtonWidgetProps {
							text: String::from("+"),
							font_size: 24.0,
							..Default::default()
						}}
						on_event={on_button_click!(ResMut<AudioSettings>, |mut settings: ResMut<
							AudioSettings,
						>| {
							settings.music_volume = (settings.music_volume + 0.05).min(1.0);
							save_config_file(Options {
								audio: *settings
							});
						})}
					/>
				</BackgroundBundle>

				<BackgroundBundle
					styles={row_style}
				>
					<TextWidgetBundle
						text={TextProps {
							content: String::from("UI Volume"),
							size: 32.0,
							..Default::default()
						}}
					/>
					<BackgroundBundle styles={KStyle { width: stretch(1.0), ..Default::default() }} />
					<ButtonWidget
						props={ButtonWidgetProps {
							text: String::from("-"),
							font_size: 24.0,
							..Default::default()
						}}
						on_event={on_button_click!(ResMut<AudioSettings>, |mut settings: ResMut<
							AudioSettings,
						>| {
							settings.ui_volume = (settings.ui_volume - 0.05).max(0.0);
							save_config_file(Options {
								audio: *settings
							});
						})}
					/>
					<BackgroundBundle
						styles={KStyle {
							padding_left: stretch(1.0),
							padding_right: stretch(1.0),
							padding_top: stretch(1.0),
							padding_bottom: stretch(1.0),
							width: px(100.0),
							..Default::default()
						}}
					>
						<TextWidgetBundle
							text={TextProps {
								content: format!("{:.0}", audio.ui_volume * 100.0),
								size: 32.0,
								..Default::default()
							}}
						/>
					</BackgroundBundle>
					<ButtonWidget
						props={ButtonWidgetProps {
							text: String::from("+"),
							font_size: 24.0,
							..Default::default()
						}}
						on_event={on_button_click!(ResMut<AudioSettings>, |mut settings: ResMut<
							AudioSettings,
						>| {
							settings.ui_volume = (settings.ui_volume + 0.05).min(1.0);
							save_config_file(Options {
								audio: *settings
							});
						})}
					/>
				</BackgroundBundle>
			</PanelWidget>

			<ButtonWidget
				props={ButtonWidgetProps { text: String::from("Back"), font_size: 32.0, ..Default::default() }}
				on_event={on_button_click!(Commands, |mut commands: Commands| {
					commands.insert_resource(NextState(AppState::Menu));
				})}
			/>
		</ElementBundle>

	};

	true
}
