use bevy::prelude::*;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	BackgroundBundle, ElementBundle, ScrollBoxBundle, ScrollBoxProps, ScrollContextProviderBundle,
	TextProps, TextWidgetBundle,
};

use crate::states::Player;
use crate::ui::components::*;
use crate::ui::prelude::*;
use crate::ui::sync::{ItemTradeData, UIStatsData, UITradeData};
use crate::ui::widgets::*;
use crate::world::{CurrentResidence, MapQuery, TownName, TownPaths, TradeGood, TradingState};
use crate::{basic_widget, empty_props, on_button_click};

empty_props!(ShopPanelProps);
basic_widget!(ShopPanelProps => ShopPanel);

pub fn buysell_button_factory(target: String, is_buy: bool) -> OnEvent {
	let target = target.clone();
	on_button_click!(
		ParamSet<(Query<&mut TradingState>, Res<UITradeData>,)>,
		|mut params: ParamSet<(Query<&mut TradingState>, Res<UITradeData>,)>| {
			let player_entity = params.p1().player_entity.clone();
			let town_entity = params.p1().town_entity.clone();
			let entry = {
				let shop_items = &params.p1().shop_items;
				shop_items
					.iter()
					.find(|en| en.definition.name == target)
					.cloned()
			};

			if let (Some(player), Some(town), Some(entry)) = (player_entity, town_entity, entry) {
				let mut query = params.p0();

				if query.contains(player) && query.contains(town) {
					let mut has_purchased = false;
					if is_buy {
						if let Ok(mut player_inv) = query.get_mut(player) {
							has_purchased =
								player_inv.try_buy_items(entry.current_cost, target.clone(), 1);
						}
						if has_purchased {
							if let Ok(mut town_inv) = query.get_mut(town) {
								town_inv.try_sell_items(entry.current_cost, target.clone(), 1);
							}
						}
					} else {
						let adjusted_sell_price = adjust_player_sell_price(entry.current_cost);
						if let Ok(mut town_inv) = query.get_mut(town) {
							has_purchased =
								town_inv.try_buy_items(adjusted_sell_price, target.clone(), 1);
						}
						if has_purchased {
							if let Ok(mut player_inv) = query.get_mut(player) {
								player_inv.try_sell_items(adjusted_sell_price, target.clone(), 1);
							}
						}
					}
				}
			}
		}
	)
}

fn adjust_player_sell_price(base: usize) -> usize {
	(base as f32 * 0.8).ceil() as usize
}

pub fn render_shop_panel(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	query: Query<&ShopPanelProps>,
	ui_data: Res<UITradeData>,
) -> bool {
	let parent_id = Some(entity);

	if ui_data.is_starved {
		rsx! {
			<ElementBundle>
				<TextWidgetBundle
					text={TextProps {
						content: "As you wander into the merchant's store, you find only ".to_string(),
						size: 32.0,
						..Default::default()
					}}
					styles={KStyle {
						color: value(Color::BLACK),
						padding: edge_px(20.0),
						top: stretch(1.0),
						bottom: stretch(1.0),
						left: stretch(1.0),
						right: stretch(1.0),
						..Default::default()
					}}
				/>
			</ElementBundle>
		};
	} else if let Ok(props) = query.get(entity) {
		rsx! {
			<ElementBundle>
				<TextWidgetBundle
					text={TextProps {
						content: format!("You find an array of goods to buy:"),
						size: 32.0,
						..Default::default()
					}}
					styles={KStyle {
						color: value(Color::BLACK),
						padding: edge_px(20.0),
						bottom: px(15.0),
						left: stretch(1.0),
						right: stretch(1.0),
						..Default::default()
					}}
				/>
				<ScrollContextProviderBundle>
					<ScrollBoxBundle
						scroll_box_props={ScrollBoxProps {
							track_color: Some(Color::rgb(0.827, 0.482, 0.353)),
							thumb_color: Some(Color::rgb(0.451, 0.224, 0.063)),
							..Default::default()
						} } styles={KStyle { pointer_events: value(PointerEvents::None), padding: edge_px(5.0), ..Default::default() }}>
						{
								for ItemTradeData {
									town_amount,
									player_amount,
									current_cost,
									definition: TradeGood { name, icon, .. },
								} in &ui_data.shop_items
								{
									constructor! {
										<BackgroundBundle
											styles={KStyle {
												pointer_events: value(PointerEvents::ChildrenOnly),
												layout_type: value(LayoutType::Row),
												col_between: stretch(0.5),
												height: px(40.0),
												bottom: px(15.0),
												padding_top: px(4.0),
												padding_bottom: px(4.0),
												color: value(Color::BLACK),
												..Default::default()
											}}
										>
											<InsetIconWidget props={InsetIconProps { image: icon.clone(), size: 32.0}} />
											<TextWidgetBundle styles={KStyle { width: stretch(1.75), ..Default::default() }} text={TextProps { content: name.clone(), size: 30.0, line_height: Some(32.0), ..Default::default() } }/>
											<TextWidgetBundle styles={KStyle { width: stretch(0.6), ..Default::default() }} text={TextProps { content: format!("{}g", current_cost), size: 30.0, line_height: Some(32.0), ..Default::default() } }/>
											<TextWidgetBundle styles={KStyle { width: stretch(0.5), ..Default::default() }} text={TextProps { content: format!("{}", town_amount), size: 30.0, line_height: Some(32.0), ..Default::default() } }/>
											<TextWidgetBundle styles={KStyle { width: stretch(0.5), ..Default::default() }} text={TextProps { content: format!("{}", player_amount), size: 30.0, line_height: Some(32.0), ..Default::default() } }/>
											<ButtonWidget on_event={buysell_button_factory(name.clone(), true)} props={ButtonWidgetProps { font_size: 20.0, is_disabled: ui_data.player_gold < *current_cost || town_amount == &0, text: String::from("Buy"), ..Default::default() }} />
											<ButtonWidget on_event={buysell_button_factory(name.clone(), false)} props={ButtonWidgetProps { font_size: 20.0, is_disabled: ui_data.town_gold < adjust_player_sell_price(*current_cost) || player_amount == &0, text: String::from("Sell"), ..Default::default() }} />
										</BackgroundBundle>
									}
								}
						}
					</ScrollBoxBundle>
				</ScrollContextProviderBundle>
			</ElementBundle>
		};
	}

	true
}
