use bevy::prelude::*;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	ElementBundle, ScrollBoxBundle, ScrollBoxProps, ScrollContextProviderBundle, TextProps,
	TextWidgetBundle,
};

use crate::assets::AssetHandles;
use crate::states::Player;
use crate::system::utilities::format_ui_distance;
use crate::ui::components::*;
use crate::ui::prelude::*;
use crate::ui::sync::{UITradeData, UITravelInfo};
use crate::ui::widgets::*;
use crate::world::{
	CraftSpecialism, CurrentResidence, MapQuery, TownName, TownPaths, TradingState,
};
use crate::{basic_widget, empty_props, on_button_click};

empty_props!(TavernPanelProps);
basic_widget!(TavernPanelProps => TavernPanel);

#[derive(Eq, Default, PartialEq, Debug, Clone, Component)]
pub struct TavernPanelState {
	pub hint: Option<String>,
}

pub fn render_tavern_panel(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	ui_data: Res<UITradeData>,
	query: Query<&TavernPanelProps>,
	state_query: Query<&TavernPanelState>,
	foo: Query<(&TownName, &CraftSpecialism)>,
) -> bool {
	let parent_id = Some(entity);
	let state_entity = widget_context.use_state(&mut commands, entity, TavernPanelState::default());

	let buy_hint = on_button_click!(
		ParamSet<(
			Query<&mut TradingState, With<Player>>,
			Query<(&TownName, &CraftSpecialism)>,
			Query<&mut TavernPanelState>
		)>,
		|mut params: ParamSet<(
			Query<&mut TradingState, With<Player>>,
			Query<(&TownName, &CraftSpecialism)>,
			Query<&mut TavernPanelState>
		)>| {
			let success = if let Ok(mut trade_state) = params.p0().get_single_mut() {
				trade_state.spend_gold(50)
			} else {
				false
			};

			let rumour = if success {
				let query = params.p1();
				let mut viable_iter = query.iter();
				let viable_count = viable_iter.len();
				if viable_count < 1 {
					None
				} else {
					let selected = viable_iter.nth(fastrand::usize(0..viable_count));
					selected.map(|(name, craft)| craft.format_rumour(&name.0))
				}
			} else {
				None
			};

			if let Ok(mut state_state) = params.p2().get_mut(state_entity) {
				state_state.hint = rumour.clone();
			}
		}
	);

	if let (Ok(props), Ok(state)) = (query.get(entity), state_query.get(state_entity)) {
		rsx! {
			<ElementBundle>
				<TextWidgetBundle
					styles={KStyle {
						left: stretch(1.0),
						right: stretch(1.0),
						color: value(Color::BLACK),
						..Default::default()
					}}
					text={TextProps {
						size: 32.0,
						content: if ui_data.is_starved { String::from("The corpse of a bartender slumps over the bar") } else { String::from("A strange bartender eyes you suspiciously") },
						..Default::default()
					}}
				/>
				{ if let Some(hint) = &state.hint {
					constructor! {
						<TextWidgetBundle
							styles={KStyle {
								top: px(20.0),
								bottom: px(20.0),
								left: stretch(1.0),
								right: stretch(1.0),
								color: value(Color::BLACK),
								..Default::default()
							}}
							text={TextProps {
								size: 28.0,
								content: format!("\"{}\"", &hint),
								..Default::default()
							}}
						/>
					}
				}}


				<ScrollContextProviderBundle>
					<ScrollBoxBundle>
			{ if state.hint.is_none() {
				constructor! {
						<TextWidgetBundle
							styles={KStyle {
								top: px(20.0),
								bottom: px(20.0),
								left: stretch(1.0),
								right: stretch(1.0),
								color: value(Color::BLACK),
								..Default::default()
							}}
							text={TextProps {
								size: 28.0,
								content: ui_data.hunger_descriptor.clone(),
								..Default::default()
							}}
						/>
				}
			}}
					</ScrollBoxBundle>
				</ScrollContextProviderBundle>

				{if !ui_data.is_starved {
					constructor! {
						<ButtonWidget
							props={ButtonWidgetProps {
								font_size: 24.0,
								left_icon: IconContent::Atlas(String::from("icons"), 8),
								text: String::from("50g: Tip Barkeep"),
								is_disabled: ui_data.player_gold < 50,
								..Default::default()
							}}
							styles={KStyle {
								left: stretch(1.0),
								right: stretch(1.0),
								..Default::default()
							}}
							on_event={buy_hint}
						/>
					}
				}}
			</ElementBundle>
		};
	}

	true
}
