use bevy::prelude::*;
use iyes_loopless::prelude::NextState;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	BackgroundBundle, ElementBundle, KImage, KImageBundle, TextProps, TextWidgetBundle,
};

use crate::assets::AssetHandles;
use crate::persistance::{fs_utils, LoadFileEvent};
use crate::system::flow::AppState;
use crate::ui::components::*;
use crate::ui::context::{MainMenuContext, MainMenuView};
use crate::ui::prelude::{edge_px, pct, px, stretch, value};
use crate::ui::widgets::*;
use crate::{basic_widget, empty_props, on_button_click};

empty_props!(TipsPanelProps);
basic_widget!(TipsPanelProps => TipsPanel);

#[derive(Default, Component, Clone, Copy, Eq, PartialEq)]
pub struct TipsPanelState {
	tip_index: usize,
}

pub fn render_tips_panel(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	assets: Res<AssetHandles>,
	mut props_query: Query<&mut ComputedStyles>,
	state_query: Query<&TipsPanelState>,
) -> bool {
	let parent_id = Some(entity);

	let state_entity = widget_context.use_state(&mut commands, entity, TipsPanelState::default());

	if let Ok(mut computed) = props_query.get_mut(entity) {
		if let Ok(state) = state_query.get(state_entity) {
			let tips = vec![
				"Town merchants will only pay 80% of their sell price when buying goods from you",
				"Be sure to tip your barkeep, they usually have great information on the various towns",
				"Your game will be saved when you enter a town - make sure you reach your next destination to save your progress!",
				"Without a regular supply of foods, towns might starve. Visit the tavern to check up on the locals.",
			];

			let tips_len = tips.len();

			let contents_container_style = KStyle {
				position_type: value(KPositionType::SelfDirected),
				width: pct(65.0),
				height: pct(80.0),
				min_width: px(400.0),
				min_height: px(300.0),
				max_width: px(500.0),
				max_height: px(400.0),
				top: stretch(1.0),
				left: stretch(1.0),
				right: stretch(1.0),
				bottom: stretch(1.0),
				layout_type: value(LayoutType::Column),
				row_between: px(20.0),
				color: value(Color::BLACK),
				..Default::default()
			};

			let header_style = KStyle {
				left: stretch(1.0),
				right: stretch(1.0),
				..Default::default()
			};

			let tip_style = KStyle {
				height: stretch(1.0),
				padding_left: px(20.0),
				padding_right: px(20.0),
				..Default::default()
			};

			let button_style = KStyle {
				..Default::default()
			};

			let controls_style = KStyle {
				// top: stretch(1.0),
				height: px(60.0),
				padding_top: stretch(1.0),
				padding_bottom: stretch(1.0),
				layout_type: value(LayoutType::Row),
				col_between: stretch(1.0),
				..Default::default()
			};

			let on_next_tip = on_button_click!(Query<&mut TipsPanelState>, |mut query: Query<
				&mut TipsPanelState,
			>| {
				if let Ok(mut state) = query.get_mut(state_entity) {
					state.tip_index = if state.tip_index == tips_len - 1 {
						0
					} else {
						state.tip_index + 1
					};
				}
			});
			let on_previous_tip =
				on_button_click!(Query<&mut TipsPanelState>, |mut query: Query<
					&mut TipsPanelState,
				>| {
					if let Ok(mut state) = query.get_mut(state_entity) {
						state.tip_index = if state.tip_index == 0 {
							tips_len - 1
						} else {
							state.tip_index - 1
						};
					}
				});

			let on_back = on_button_click!(Commands, |mut commands: Commands| commands
				.insert_resource(NextState(AppState::Menu)));

			rsx! {
				<ElementBundle styles={contents_container_style}>
					<PanelWidget>
						<TextWidgetBundle
							text={TextProps {
								content: format!("Tip #{}", state.tip_index + 1),
								font: Some(String::from("header")),
								size: 52.0,
								..Default::default()
							}}
							styles={header_style}
						/>
						<BackgroundBundle
							styles={KStyle {
								width: pct(80.0),
								left: stretch(1.0),
								right: stretch(1.0),
								top: px(30.0),
								bottom: px(30.0),
								..Default::default()
							}}
						>
							<TextWidgetBundle
								text={TextProps {
									content: tips[state.tip_index].to_string(),
									size: 32.0,
									..Default::default()
								}}
							/>
						</BackgroundBundle>
					</PanelWidget>
					<BackgroundBundle styles={controls_style}>
						<ButtonWidget
							styles={button_style.clone()}
							props={ButtonWidgetProps {
								text: String::from("<"),
								font_size: 32.0,
								..Default::default()
							}}
							on_event={on_previous_tip}
						/>

						<ButtonWidget
							styles={button_style.clone()}
							props={ButtonWidgetProps {
								text: String::from("back"),
								font_size: 32.0,
								..Default::default()
							}}
							on_event={on_back}
						/>

						<ButtonWidget
							styles={button_style.clone()}
							props={ButtonWidgetProps {
								text: String::from(">"),
								font_size: 32.0,
								..Default::default()
							}}
							on_event={on_next_tip}
						/>
					</BackgroundBundle>
				</ElementBundle>

			};
		}
	}
	true
}
