use bevy::prelude::*;
use kayak_ui::prelude::*;
use kayak_ui::widgets::{
	ElementBundle, ScrollBoxBundle, ScrollBoxProps, ScrollContextProviderBundle, TextProps,
	TextWidgetBundle,
};

use crate::assets::AssetHandles;
use crate::states::Player;
use crate::ui::components::*;
use crate::ui::prelude::*;
use crate::ui::sync::{UIStatsData, UITravelInfo};
use crate::ui::widgets::*;
use crate::world::{CurrentResidence, MapQuery, TownPaths};
use crate::{basic_widget, empty_props, on_button_click};

empty_props!(TownMenuPanelProps);
basic_widget!(TownMenuPanelProps => TownMenuPanel);

#[derive(Debug, Ord, PartialOrd, PartialEq, Eq, Copy, Clone, Default)]
pub enum TownMenuTab {
	#[default]
	Travel,
	Merchant,
	Tavern,
	None,
}

#[derive(Component, Clone, PartialEq, Default)]
pub struct TownMenuPanelState {
	pub tab: TownMenuTab,
	pub next_tab: Option<TownMenuTab>,
}

pub fn render_town_menu_panel(
	In((widget_context, entity)): In<(KayakWidgetContext, Entity)>,
	mut commands: Commands,
	mut state_query: Query<&mut TownMenuPanelState>,
	ui_data: Res<UITravelInfo>,
	stats_data: Res<UIStatsData>,
	places: Res<TownPaths>,
	mut next_state: Local<TownMenuTab>,
) -> bool {
	let parent_id = Some(entity);
	let state_entity =
		widget_context.use_state(&mut commands, entity, TownMenuPanelState::default());

	let distance_style = KStyle {
		position_type: value(KPositionType::SelfDirected),
		top: stretch(1.0),
		left: stretch(1.0),
		right: stretch(1.0),
		bottom: px(50.0),
		..Default::default()
	};

	let panel_style = KStyle {
		position_type: value(KPositionType::SelfDirected),
		width: pct(65.0),
		height: pct(80.0),
		min_width: px(400.0),
		min_height: px(300.0),
		max_width: px(850.0),
		max_height: px(550.0),
		top: stretch(1.0),
		left: stretch(1.0),
		right: stretch(1.0),
		bottom: stretch(1.0),
		layout_type: value(LayoutType::Column),
		row_between: px(20.0),
		..Default::default()
	};

	let click_tab_travel = on_button_click!(Query<&mut TownMenuPanelState>, |mut q: Query<
		&mut TownMenuPanelState,
	>| {
		if let Ok(mut state) = q.get_mut(state_entity) {
			state.tab = TownMenuTab::None;
			state.next_tab = Some(TownMenuTab::Travel);
		}
	});
	let click_tab_merchant = on_button_click!(Query<&mut TownMenuPanelState>, |mut q: Query<
		&mut TownMenuPanelState,
	>| {
		if let Ok(mut state) = q.get_mut(state_entity) {
			state.tab = TownMenuTab::None;
			state.next_tab = Some(TownMenuTab::Merchant);
		}
	});
	let click_tab_tavern = on_button_click!(Query<&mut TownMenuPanelState>, |mut q: Query<
		&mut TownMenuPanelState,
	>| {
		if let Ok(mut state) = q.get_mut(state_entity) {
			state.tab = TownMenuTab::None;
			state.next_tab = Some(TownMenuTab::Tavern);
		}
	});

	if let (Some(ref place), Some(mut state)) = (
		&ui_data.current_town,
		state_query.get_mut(state_entity).ok(),
	) {
		rsx! {
			<ElementBundle styles={panel_style}>
				<PanelWidget>
					<TextWidgetBundle
						text={TextProps {
							font: Some(String::from("header")),
							content: format!("Hark! You enter {}", &place),
							size: 48.0,
							..Default::default()
						}}
						styles={KStyle {
							color: value(Color::BLACK),
							padding: edge_px(20.0),
							left: stretch(1.0),
							right: stretch(1.0),
							..Default::default()
						}}
					/>

					<VDividerWidget props={VDividerWidgetProps { height: 4.0, padding: 5.0, color: Color::rgb(0.52, 0.369, 0.18)}} />

					<ElementBundle styles={KStyle {
						layout_type: value(LayoutType::Row),
						padding_top: stretch(1.0),
						padding_bottom: stretch(1.0),
						padding_left: stretch(1.0),
						padding_right: stretch(1.0),
						height: px(60.0),
						col_between: px(15.0),
						..Default::default()
					}}>
						<ButtonWidget
							props={ButtonWidgetProps {
								left_icon: IconContent::Atlas(String::from("icons"), 4),
								..ButtonWidgetProps::text("Travel", 28.0)
							}}
							on_event={click_tab_travel}
						/>
						<ButtonWidget
							props={ButtonWidgetProps {
								left_icon: IconContent::Atlas(String::from("icons"), 9),
								..ButtonWidgetProps::text("Merchant", 28.0)
							}}
							on_event={click_tab_merchant}
						/>
						<ButtonWidget
							props={ButtonWidgetProps {
								left_icon: IconContent::Atlas(String::from("icons"), 11),
								..ButtonWidgetProps::text("Tavern", 28.0)
							}}
							on_event={click_tab_tavern}
						/>
					</ElementBundle>

					<VDividerWidget props={VDividerWidgetProps { height: 4.0, padding: 5.0, color: Color::rgb(0.52, 0.369, 0.18)}} />


					{
						match state.tab {
							TownMenuTab::Travel => {
								constructor! {
									<TransitPanel />
								};
							},
							TownMenuTab::Merchant => {
								constructor! {
									<ShopPanel />
								}
							}
							TownMenuTab::Tavern => {
								constructor! {
									<TavernPanel />
								}
							}
							TownMenuTab::None => {}
						}
					}
				</PanelWidget>

				<PanelWidget styles={KStyle { layout_type: value(LayoutType::Row), height: px(45.0),  ..Default::default() }}>
					<PanelStatWidget
						styles={KStyle {
							position_type: value(KPositionType::SelfDirected),
							left: stretch(1.0),
							top: stretch(0.8),
							bottom: stretch(1.0),
							right: px(40.0),
							..Default::default()
						}}
						props={PanelStatProps {
							fill: false,
							text: format!("{}g", stats_data.money),
							font_size: 24.0,
							icon: IconContent::Atlas(String::from("icons"), 8),
							color: Color::BLACK
						}}
					/>
				</PanelWidget>
			</ElementBundle>

		};

		if let Some(next) = state.next_tab {
			state.tab = next;
			state.next_tab = None;
		}
	}

	true
}
