use bevy::math::Vec3Swizzles;
use bevy::prelude::*;
use fake::faker::company::en::{Bs, BsAdj, BsNoun, BsVerb};
use fake::faker::lorem::en::{Paragraph, Paragraphs};
use fake::Fake;
use ldtk_rust::{EntityInstance, FieldInstance};
use num_traits::AsPrimitive;
use serde::{Deserialize, Serialize};
use serde_json::Value;

use crate::assets::{AssetHandles, EncounterSet};
use crate::states::Player;
use crate::world::travel::WorldTickEvent;
use crate::world::utils::entity_to_worldspace;
use crate::world::{HungerState, TradingState};

#[derive(Copy, Clone, PartialEq, Debug)]
pub enum EncounterType {
	MildField,
	Bridge,
	Monastery,
	Docks,
	Desert,
	GrassyRoad,
}

impl EncounterType {
	pub fn get_custom_encounters(&self) -> Option<Vec<Encounter>> {
		match self {
			Self::Docks => Some(vec![Encounter {
				title: String::from("Your journey begins"),
				description: String::from(include_str!("../assets/static/intro_text.txt")),
				options: vec![EncounterOption {
					requirements: Vec::new(),
					label: String::from("Journey Forth"),
					outcomes: vec![],
					hide_missing_reqs: true,
				}],
			}]),
			_ => None,
		}
	}
	pub fn get_asset_name(&self) -> String {
		String::from("grasslands")
	}
}

impl TryFrom<String> for EncounterType {
	type Error = &'static str;

	fn try_from(value: String) -> Result<Self, Self::Error> {
		match &*value {
			"MildField" => Ok(Self::MildField),
			"Bridge" => Ok(Self::Bridge),
			"Monastery" => Ok(Self::Monastery),
			"Docks" => Ok(Self::Docks),
			"Desert" => Ok(Self::Desert),
			"GrassyRoad" => Ok(Self::GrassyRoad),
			_ => Err("Invalid encounter type"),
		}
	}
}

#[derive(Copy, Clone, PartialEq, Debug)]
pub struct EncounterZone {
	pub zone_type: EncounterType,
	pub area: Rect,
}

impl EncounterZone {
	pub fn new(
		zone_type: EncounterType,
		cx: impl AsPrimitive<f32>,
		cy: impl AsPrimitive<f32>,
		width: impl AsPrimitive<f32>,
		height: impl AsPrimitive<f32>,
	) -> Self {
		let cx = cx.as_();
		let cy = cy.as_();
		let width = width.as_();
		let height = height.as_();

		EncounterZone {
			zone_type,
			area: Rect::new(
				cx - (width / 2.0),
				cy - (height / 2.0),
				cx + (width / 2.0),
				cy + (height / 2.0),
			),
		}
	}

	pub fn contains(&self, point: Vec2) -> bool {
		self.area.contains(point)
	}
}

#[derive(Default, Resource, Debug)]
pub struct WorldZones(pub Vec<EncounterZone>);

impl WorldZones {
	pub fn get_container(&self, point: Vec2) -> Option<EncounterZone> {
		self.0.iter().find(|e| e.contains(point)).copied()
	}
	pub fn from_entities(height: i64, value: Vec<&EntityInstance>) -> Self {
		value
			.iter()
			.filter_map(|instance| {
				if instance.identifier == String::from("DangerZone") {
					let (cx, cy) = entity_to_worldspace(height, instance);
					match instance
						.field_instances
						.iter()
						.find(|field| field.identifier == String::from("zone_type"))
					{
						Some(v) => match &v.value {
							Some(Value::String(zone_type)) => {
								EncounterType::try_from(zone_type.clone()).ok().map(|ze| {
									EncounterZone::new(ze, cx, cy, instance.width, instance.height)
								})
							}
							_ => None,
						},
						_ => None,
					}
				} else {
					None
				}
			})
			.collect::<Vec<EncounterZone>>()
			.into()
	}
}

impl From<Vec<EncounterZone>> for WorldZones {
	fn from(value: Vec<EncounterZone>) -> Self {
		Self(value)
	}
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(untagged)]
pub enum RewardAmount {
	Exact(usize),
	Between(usize, usize),
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(tag = "type")]
pub enum EncounterOutcome {
	Continue,
	Description {
		text: String,
	},
	GainResource {
		resource_type: String,
		amount: RewardAmount,
	},
	GainSustenance {
		amount: RewardAmount,
	},
	LoseSustenance {
		amount: RewardAmount,
	},
	GainGold {
		amount: RewardAmount,
	},
	LoseGold {
		amount: RewardAmount,
	},
	LoseResource {
		resource_type: String,
		amount: RewardAmount,
	},
	TickWorld {
		amount: RewardAmount,
	},
	Ambush {
		description: String,
		strength: usize,
		defence: usize,
		victory: Vec<EncounterOutcome>,
		defeat: Vec<EncounterOutcome>,
	},
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct ChanceOutcome {
	pub chance: usize,
	pub outcome: EncounterOutcome,
}

#[derive(Clone, Debug, Serialize, Deserialize)]
#[serde(tag = "type")]
pub enum EncounterRequirement {
	MinimumItemAmount { item: String, amount: usize },
	MaximumItemAmount { item: String, amount: usize },
	MinimumGoldAmount { amount: usize },
	MaximumGoldAmount { amount: usize },
	AnyFoodLessThan { amount: usize, or_equal: bool },
	AnyFoodGreaterThan { amount: usize, or_equal: bool },
}

fn default_false() -> bool {
	false
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct EncounterOption {
	pub label: String,
	pub outcomes: Vec<EncounterOutcome>,
	#[serde(default = "Vec::new")]
	pub requirements: Vec<EncounterRequirement>,
	#[serde(default = "default_false")]
	pub hide_missing_reqs: bool,
}

impl EncounterOption {
	pub fn describe(&self) -> String {
		let mut description = Vec::with_capacity(self.outcomes.len());
		for outcome in &self.outcomes {
			match outcome {
				EncounterOutcome::Continue => {}
				EncounterOutcome::Description { text } => {
					description.push(text.clone());
				}
				EncounterOutcome::GainResource { resource_type, .. } => {
					description.push(format!("Gained {}", resource_type));
				}
				EncounterOutcome::GainSustenance { .. } => {
					description.push(String::from("Your Hunger Has Decreased"));
				}
				EncounterOutcome::LoseSustenance { .. } => {
					description.push(String::from("Your Hunger Has Decreased"));
				}
				EncounterOutcome::GainGold { .. } => {
					description.push(String::from("Gained Some Gold"));
				}
				EncounterOutcome::LoseGold { .. } => {
					description.push(String::from("Lost Some Gold"));
				}
				EncounterOutcome::LoseResource { resource_type, .. } => {
					description.push(format!("Gained {}", resource_type));
				}
				EncounterOutcome::TickWorld { .. } => {
					description.push(String::from("Some time has passed"))
				}
				EncounterOutcome::Ambush { .. } => {}
			}
		}

		description.join("\n")
	}
}

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Encounter {
	pub title: String,
	pub description: String,
	pub options: Vec<EncounterOption>,
}

pub fn gen_encounter() -> Encounter {
	Encounter {
		title: Bs().fake(),
		description: Paragraph(1..3).fake(),
		options: vec![
			EncounterOption {
				label: BsVerb().fake(),
				outcomes: vec![EncounterOutcome::GainResource {
					resource_type: format!(
						"{} {}",
						BsAdj().fake::<String>(),
						BsNoun().fake::<String>()
					),
					amount: {
						let b = fastrand::usize(2..5);
						let inc = fastrand::usize(2..5);
						RewardAmount::Between(b, b + inc)
					},
				}],
				requirements: Vec::new(),
				hide_missing_reqs: true,
			},
			EncounterOption {
				label: BsVerb().fake(),
				outcomes: vec![EncounterOutcome::LoseResource {
					resource_type: format!(
						"{} {}",
						BsAdj().fake::<String>(),
						BsNoun().fake::<String>()
					),
					amount: {
						let b = fastrand::usize(2..5);
						let inc = fastrand::usize(2..5);
						RewardAmount::Between(b, b + inc)
					},
				}],
				requirements: Vec::new(),
				hide_missing_reqs: true,
			},
		],
	}
}

#[derive(Clone, Default, Resource, Debug, Serialize, Deserialize)]
pub enum EncounterState {
	#[default]
	NoEncounter,
	Choice(Encounter),
	Consequence(Encounter, EncounterOption),
}

impl EncounterState {
	pub fn is_in_encounter(&self) -> bool {
		!matches!(self, Self::NoEncounter)
	}
}

#[derive(Clone, Debug)]
pub struct SelectEncounterOptionEvent {
	pub option: EncounterOption,
}

pub fn handle_encounter_option_event(
	mut events: ResMut<Events<SelectEncounterOptionEvent>>,
	mut time_ticks: EventWriter<WorldTickEvent>,
	mut player_query: Query<(&mut TradingState, &mut HungerState), With<Player>>,
	mut encounter_state: ResMut<EncounterState>,
) {
	for event in events.drain() {
		for (mut trading, mut hunger) in &mut player_query {
			for outcome in &event.option.outcomes {
				log::info!("Processing outcome: {:?}", outcome);

				match outcome {
					EncounterOutcome::Continue => {
						*encounter_state = EncounterState::NoEncounter;
					}
					EncounterOutcome::Description { .. } => {}
					EncounterOutcome::GainResource {
						resource_type,
						amount,
					} => match amount {
						RewardAmount::Exact(number) => {
							trading.add_items(resource_type, *number);
						}
						RewardAmount::Between(min, max) => {
							trading.add_items(resource_type, fastrand::usize(*min..*max));
						}
					},
					EncounterOutcome::LoseResource {
						resource_type,
						amount,
					} => match amount {
						RewardAmount::Exact(number) => {
							trading.force_remove_items(resource_type, *number);
						}
						RewardAmount::Between(min, max) => {
							trading.force_remove_items(resource_type, fastrand::usize(*min..*max));
						}
					},
					EncounterOutcome::GainSustenance { amount } => match amount {
						RewardAmount::Exact(number) => {
							hunger.sustenance += *number;
						}
						RewardAmount::Between(min, max) => {
							hunger.sustenance += fastrand::usize(*min..*max);
						}
					},
					EncounterOutcome::LoseSustenance { amount } => match amount {
						RewardAmount::Exact(number) => {
							hunger.sustenance = hunger.sustenance.saturating_sub(*number);
						}
						RewardAmount::Between(min, max) => {
							hunger.sustenance = hunger
								.sustenance
								.saturating_sub(fastrand::usize(*min..*max));
						}
					},
					EncounterOutcome::GainGold { amount } => match amount {
						RewardAmount::Exact(number) => {
							trading.adjust_gold(*number);
						}
						RewardAmount::Between(min, max) => {
							trading.adjust_gold(fastrand::usize(*min..*max));
						}
					},
					EncounterOutcome::LoseGold { amount } => match amount {
						RewardAmount::Exact(number) => {
							trading.adjust_gold(-(*number as isize));
						}
						RewardAmount::Between(min, max) => {
							trading.adjust_gold(-(fastrand::usize(*min..*max) as isize));
						}
					},
					EncounterOutcome::TickWorld { amount } => {
						// match amount {
						// RewardAmount::Exact(number) => {
						// 	for _ in 0..*number {
						// 		time_ticks.send(WorldTickEvent);
						// 	}
						// }
						// RewardAmount::Between(min, max) => {
						// 	for _ in 0..fastrand::usize(*min..*max) {
						// 		time_ticks.send(WorldTickEvent);
						// 	}
						// }
						// }
					}
					EncounterOutcome::Ambush { .. } => {}
				}
			}
		}

		if let EncounterState::Choice(encounter) = &*encounter_state {
			*encounter_state = EncounterState::Consequence(encounter.clone(), event.option.clone());
		}
		log::info!("Finished processing");
	}
}

pub fn notify_new_zone(
	zones: Res<WorldZones>,
	player_query: Query<&Transform, With<Player>>,
	mut last_zone: Local<Option<EncounterZone>>,
	mut encounter_state: ResMut<EncounterState>,
	assets: Res<AssetHandles>,
	sets: Res<Assets<EncounterSet>>,
) {
	for position in &player_query {
		if let Some(zone) = zones.get_container(position.translation.xy()) {
			match *last_zone {
				Some(existing_zone) => {
					if zone != existing_zone {
						match zone.zone_type.get_custom_encounters() {
							Some(envc) => {
								log::info!("Recounter Zone: {:?}", zone.zone_type);
								let encounter = envc[fastrand::usize(0..envc.len())].clone();
								*last_zone = Some(zone);
								*encounter_state = EncounterState::Choice(encounter);
							}
							None => {
								let asset_name = zone.zone_type.get_asset_name();
								let handle = assets.encounter_set(asset_name);
								let set = sets.get(&handle).unwrap();

								log::info!("New Zone: {:?}", zone.zone_type);
								*last_zone = Some(zone);
								*encounter_state = EncounterState::Choice(set.select());
							}
						}
					}
				}
				None => match zone.zone_type.get_custom_encounters() {
					Some(envc) => {
						log::info!("Recounter Zone: {:?}", zone.zone_type);
						let encounter = envc[fastrand::usize(0..envc.len())].clone();
						*last_zone = Some(zone);
						*encounter_state = EncounterState::Choice(encounter);
					}
					None => {
						let asset_name = zone.zone_type.get_asset_name();
						let handle = assets.encounter_set(asset_name);
						let set = sets.get(&handle).unwrap();

						log::info!("New Zone: {:?}", zone.zone_type);
						*last_zone = Some(zone);
						*encounter_state = EncounterState::Choice(set.select());
					}
				},
			}
		} else if last_zone.is_some() {
			*last_zone = None;
		}
	}
}
