use std::env::var;

use bevy::math::{uvec2, UVec2};
use fastrand::Rng;
use noise::{Blend, Clamp, NoiseFn, Perlin, Simplex};

use crate::system::utilities::Indexer;

pub struct Overworld {
	pub width: usize,
	pub height: usize,
	pub terrain: Vec<WorldTileType>,
}

#[derive(Debug, Copy, Clone, PartialOrd, PartialEq, Ord, Eq)]
#[repr(C)]
pub enum WorldTileType {
	Water = 0,
	Sand = 1,
	Dirt = 2,
	Grass = 3,
	Mountain = 4,
	Swamp = 5,
	Snow = 6,
	Desert = 7,
}

impl WorldTileType {
	pub fn get_basic_tile(&self) -> usize {
		match self {
			WorldTileType::Water => 225,
			WorldTileType::Sand => vec![294, 295, 296, 297, 298][fastrand::usize(0..5)],
			WorldTileType::Dirt => vec![1, 2, 3, 4, 5][fastrand::usize(0..5)],
			WorldTileType::Grass => 0,
			WorldTileType::Mountain => 68, // vec![68, 6, 7, 8, 9, 10][fastrand::usize(0..6)],
			WorldTileType::Swamp => vec![43, 44, 45][fastrand::usize(0..3)],
			WorldTileType::Snow => vec![11, 12, 13, 14][fastrand::usize(0..4)],
			WorldTileType::Desert => vec![326, 327, 328, 329, 330][fastrand::usize(0..5)],
		}
	}
}

fn cap_variance(var: &mut usize) {
	*var = (*var).clamp(35, 65);
}

fn adjust_variance(var: &mut usize, rng: &Rng) {
	if (rng.f32() - (*var as f32 / 100.0)) > 0.3 {
		*var = var.saturating_add(rng.usize(0..2));
	} else {
		*var = var.saturating_sub(rng.usize(0..2));
	}
}

fn random_ground_spot(indexer: &Indexer, values: &[WorldTileType], rng: &Rng) -> UVec2 {
	let width = indexer.width() as u32;
	let height = indexer.height() as u32;

	let min_x = width / 4;
	let min_y = height / 4;
	let max_x = (width / 4) * 3;
	let max_y = (height / 4) * 3;

	let mut attempts = 500;
	while attempts > 0 {
		let guess_x = rng.u32(min_x..=max_x);
		let guess_y = rng.u32(min_y..=max_y);

		if values[indexer.index(guess_x as usize, guess_y as usize)] == WorldTileType::Grass {
			return uvec2(guess_x, guess_y);
		}

		attempts -= 1;
	}

	UVec2::new(0, 0)
}

fn draw_across(
	x: usize,
	y: usize,
	width: usize,
	tile: WorldTileType,
	idx: &Indexer,
	values: &mut [WorldTileType],
) {
	let half = width / 2;
	for x_val in x - half..x + half {
		if idx.is_coordinate_valid(x_val, y) && values[idx.index(x_val, y)] == WorldTileType::Grass
		{
			values[idx.index(x_val, y)] = tile;
		}
	}
}

fn draw_updown(
	x: usize,
	y: usize,
	height: usize,
	tile: WorldTileType,
	idx: &Indexer,
	values: &mut [WorldTileType],
) {
	let half = height / 2;
	for y_val in y - half..y + half {
		if idx.is_coordinate_valid(x, y_val) && values[idx.index(x, y_val)] == WorldTileType::Grass
		{
			values[idx.index(x, y_val)] = tile;
		}
	}
}

pub fn count_world_tiles(world: &Overworld, tile_type: WorldTileType) -> usize {
	world.terrain.iter().fold(0, |count, current| {
		if *current == tile_type {
			count + 1
		} else {
			count
		}
	})
}

pub fn generate_overworld(width: usize, height: usize, rng: Rng) -> Overworld {
	let mut attempts = 100;
	let mut world = generate_overworld_data(width, height, &rng);

	loop {
		let total_tiles = width * height;
		let water_tiles = count_world_tiles(&world, WorldTileType::Water);
		let mountain_tiles = count_world_tiles(&world, WorldTileType::Mountain);
		let ground_tiles = count_world_tiles(&world, WorldTileType::Grass);

		// Regen world if it's not too good:
		// - more than 10% mountain
		// - less than 50% regular grass
		if mountain_tiles > ((total_tiles - water_tiles) / 10)
			|| ground_tiles < ((total_tiles - water_tiles) / 2)
		{
			world = generate_overworld_data(width, height, &rng);
		} else {
			return world;
		}

		attempts -= 1;
		if attempts == 0 {
			break;
		}
	}

	world
}

pub fn generate_overworld_data(width: usize, height: usize, rng: &Rng) -> Overworld {
	let indexer = Indexer::new(width, height);
	let mut values = vec![WorldTileType::Grass; width * height];

	let mut variance = 50;

	for x in 0..width {
		let mut y = variance;
		for y_val in 0..y {
			values[indexer.index(x, y_val)] = WorldTileType::Water;
		}
		for y_val in y..y + (y / 2) {
			values[indexer.index(x, y_val)] = WorldTileType::Sand;
		}

		adjust_variance(&mut variance, &rng);
	}

	cap_variance(&mut variance);

	for x in 0..width {
		let mut y = height - variance;
		for y_val in y..height {
			values[indexer.index(x, y_val)] = WorldTileType::Water;
		}
		for y_val in y - (variance / 2)..y {
			values[indexer.index(x, y_val)] = WorldTileType::Sand;
		}

		adjust_variance(&mut variance, &rng);
	}

	cap_variance(&mut variance);

	for y in 0..width {
		let mut x = variance;
		for x_val in 0..x {
			values[indexer.index(x_val, y)] = WorldTileType::Water;
		}
		for x_val in x..x + (x / 2) {
			if values[indexer.index(x_val, y)] == WorldTileType::Water {
				break;
			}
			values[indexer.index(x_val, y)] = WorldTileType::Sand;
		}

		adjust_variance(&mut variance, &rng);
	}

	cap_variance(&mut variance);

	for y in 0..width {
		let mut x = width - variance;
		for x_val in x..width {
			values[indexer.index(x_val, y)] = WorldTileType::Water;
		}
		// + (rng.usize(0..(variance / 2).max(1))))
		for x_val in x - (variance / 2)..x {
			if values[indexer.index(x_val, y)] == WorldTileType::Water {
				break;
			}
			values[indexer.index(x_val, y)] = WorldTileType::Sand;
		}

		adjust_variance(&mut variance, &rng);
	}

	let mountains = rng.usize(1..4);
	for _ in 0..mountains {
		let position = random_ground_spot(&indexer, values.as_slice(), &rng);
		let width = rng.usize(35..65);
		let height = rng.usize(35..65);
		let vertical = rng.bool();

		if vertical {
			let mut width_remaining = width;
			for y in position.y as usize..position.y as usize + (height / 3) * 2 {
				if width_remaining == 0 {
					break;
				}

				draw_across(
					position.x as usize,
					y,
					width_remaining,
					WorldTileType::Mountain,
					&indexer,
					&mut values,
				);

				width_remaining -= rng.usize(0..3);
			}

			let mut width_remaining = width;
			for y in (position.y as usize - (height / 3)..position.y as usize).rev() {
				if width_remaining == 0 {
					break;
				}

				draw_across(
					position.x as usize,
					y,
					width_remaining,
					WorldTileType::Mountain,
					&indexer,
					&mut values,
				);

				width_remaining -= rng.usize(0..4);
			}
		} else {
			let mut height_remaining = height;
			for x in position.x as usize..position.x as usize + (width / 3) * 2 {
				if height_remaining == 0 {
					break;
				}

				draw_updown(
					x,
					position.y as usize,
					height_remaining,
					WorldTileType::Mountain,
					&indexer,
					&mut values,
				);

				height_remaining -= rng.usize(0..3);
			}

			let mut height_remaining = height;
			for x in (position.x as usize - (width / 3)..position.x as usize).rev() {
				if height_remaining == 0 {
					break;
				}

				draw_updown(
					x,
					position.y as usize,
					height_remaining,
					WorldTileType::Mountain,
					&indexer,
					&mut values,
				);

				height_remaining -= rng.usize(0..4);
			}
		}
	}
	//
	let biome_noise = Blend::new(
		Perlin::new(rng.u32(0..u32::MAX)),
		Simplex::new(rng.u32(0..u32::MAX)),
		Perlin::new(rng.u32(0..u32::MAX)),
	);

	// let biome_noise = Clamp::new(Simplex::new(rng.u32(0..u32::MAX)));

	for x in 0..width {
		for y in 0..height {
			let value = biome_noise
				.get([x as f64 / width as f64, y as f64 / height as f64])
				.abs();

			if values[indexer.index(x, y)] == WorldTileType::Grass {
				if value > 0.8 {
					values[indexer.index(x, y)] = WorldTileType::Mountain;
				} else if value > 0.6 {
					values[indexer.index(x, y)] = WorldTileType::Snow;
				} else if value > 0.3 && value < 0.45 {
					values[indexer.index(x, y)] = WorldTileType::Desert;
				} else if value > 0.15 && value <= 0.2 {
					values[indexer.index(x, y)] = WorldTileType::Swamp;
				}
			}
		}
	}

	Overworld {
		terrain: values,
		width,
		height,
	}
}
