use bevy::math::vec3;
use bevy::prelude::*;
use bevy_ecs_tilemap::map::{TilemapSize, TilemapTileSize, TilemapType};
use bevy_ecs_tilemap::prelude::{TilePos, TileTextureIndex, TilemapId, TilemapTexture};
use bevy_ecs_tilemap::tiles::{TileBundle, TileStorage};
use bevy_ecs_tilemap::TilemapBundle;
use fastrand::Rng;

use crate::assets::AssetHandles;
use crate::system::camera::GameCamera;
use crate::system::utilities::Indexer;
use crate::world::generators::generate_overworld::{generate_overworld, Overworld};
use crate::world::generators::{ParticleRollingGenerator, WorldGenerator};

pub fn spawn_new_map(
	mut commands: Commands,
	assets: Res<AssetHandles>,
	mut query: Query<&mut Transform, With<GameCamera>>,
) {
	let indexer = Indexer::new(500, 500);
	let rng = Rng::new();

	let world = Overworld {
		width: 500,
		height: 500,
		terrain: ParticleRollingGenerator::new(21000, 400).generate(&indexer, &rng),
	};

	spawn_overworld(&world, &mut commands, &assets);
	for mut camera_transform in &mut query {
		camera_transform.translation =
			vec3(250.0 * 4.0, 250.0 * 4.0, camera_transform.translation.z);
	}
}

pub fn spawn_overworld(
	overworld: &Overworld,
	commands: &mut Commands,
	assets: &Res<AssetHandles>,
) -> Entity {
	let indexer = Indexer::new(overworld.width, overworld.height);
	let tilemap_size = TilemapSize {
		x: overworld.width as u32,
		y: overworld.height as u32,
	};

	let mut map_entity = commands.spawn_empty().id();
	let mut tile_storage = TileStorage::empty(tilemap_size);

	for (idx, tile) in overworld.terrain.iter().enumerate() {
		let (x, y) = indexer.reverse(idx);

		let tile_pos = TilePos {
			x: x as u32,
			y: y as u32,
		};

		let tile_entity = commands
			.spawn(TileBundle {
				position: tile_pos,
				tilemap_id: TilemapId(map_entity),
				texture_index: TileTextureIndex(tile.get_basic_tile() as u32),
				..Default::default()
			})
			.id();

		tile_storage.set(&tile_pos, tile_entity)
	}

	let tile_size = TilemapTileSize { x: 4.0, y: 4.0 };
	let grid_size = tile_size.into();
	let map_type = TilemapType::Square;

	commands.entity(map_entity).insert(TilemapBundle {
		grid_size,
		tile_size,
		size: tilemap_size,
		storage: tile_storage,
		texture: TilemapTexture::Single(assets.image("overworld")),
		map_type,
		transform: Transform::from_xyz(0.0, 0.0, 100.0),
		..Default::default()
	});

	map_entity
}
