use std::iter::zip;

use bevy::prelude::UVec2;
use fastrand::Rng;
use noise::{NoiseFn, Perlin};

use crate::system::utilities::Indexer;
use crate::world::generators::generate_overworld::WorldTileType;
use crate::world::generators::WorldGenerator;

#[derive(Copy, Clone, Debug)]
pub struct ParticleRollingGenerator {
	particle_iterations: usize,
	particle_lifespan: usize,
}

impl ParticleRollingGenerator {
	pub fn new(iterations: usize, lifespan: usize) -> Self {
		Self {
			particle_iterations: iterations,
			particle_lifespan: lifespan,
		}
	}
}

impl Default for ParticleRollingGenerator {
	fn default() -> Self {
		ParticleRollingGenerator::new(5000, 50)
	}
}

impl WorldGenerator for ParticleRollingGenerator {
	fn generate(&self, indexer: &Indexer, rng: &Rng) -> Vec<WorldTileType> {
		let mut map = self.create_map(indexer, rng);
		let mut attempts = 500;

		while attempts > 0 {
			let grass_count = map.iter().fold(0, |count, tile| {
				if *tile == WorldTileType::Grass {
					count + 1
				} else {
					count
				}
			});

			let water_count = map.iter().fold(0, |count, tile| {
				if *tile == WorldTileType::Water {
					count + 1
				} else {
					count
				}
			});

			if grass_count > (map.len() - water_count) / 2 {
				log::info!("Got good map");
				break;
			}

			log::info!("Doing a new map now; remaining {}", attempts);
			map = self.create_map(indexer, rng);
			attempts -= 1;
		}

		map
	}
}

impl ParticleRollingGenerator {
	fn create_map(&self, indexer: &Indexer, rng: &Rng) -> Vec<WorldTileType> {
		let width = indexer.width();
		let height = indexer.height();
		let mut values = vec![0.0; width * height];
		let mut mask = vec![0.0; width * height];

		for _ in 0..self.particle_iterations {
			let mut lifespan = self.particle_lifespan;
			let mut particle_x = rng.usize(width / 6..width - width / 6);
			let mut particle_y = rng.usize(height / 6..height - height / 6);

			while lifespan > 0 {
				let idx = indexer.index(particle_x, particle_y);
				let new_value = mask[idx] + 1.0 / 100.0;

				mask[idx] = new_value;
				let choices: Vec<UVec2> = indexer
					.square_adjacent(particle_x, particle_y)
					.iter()
					.filter(|UVec2 { x, y }| mask[indexer.index(*x, *y)] <= new_value)
					.copied()
					.collect();

				if !choices.is_empty() {
					let next = choices[rng.usize(0..choices.len())];
					particle_x = next.x as usize;
					particle_y = next.y as usize;
				}

				lifespan -= 1;
			}
		}

		// return mask.iter().copied().map(Self::map_noise_value).collect();

		let noise = Perlin::new(rng.u32(0..u32::MAX));
		let values: Vec<f64> = values
			.iter()
			.enumerate()
			.map(|(index, _)| {
				let (x, y) = indexer.reverse(index);
				noise.get([x as f64 / width as f64, y as f64 / height as f64])
			})
			.collect();

		zip(values, mask)
			.map(|(left, right)| Self::map_noise_value(left * right))
			.collect()
	}
}
