use std::collections::HashMap;
use std::fmt::{Debug, Formatter};
use std::io::Read;
use std::ops::{Deref, DerefMut, Index, IndexMut};

use bevy_asset::{AssetEvent, AssetLoader, Assets, BoxedFuture, Handle, LoadContext, LoadedAsset};
use bevy_ecs::component::Component;
use bevy_ecs::entity::Entity;
use bevy_ecs::event::{EventReader, EventWriter};
use bevy_ecs::system::{Commands, Query, Res, ResMut};
use bevy_reflect::TypeUuid;
use bevy_tasks::{AsyncComputeTaskPool, Task};
use futures_lite::future;
use xz2::read::XzDecoder;
// use bevy::prelude::*;

#[derive(Default)]
pub struct APackLoader;

#[derive(TypeUuid)]
#[uuid = "bce5dd7a-726c-11ed-b53f-071b386b4896"]
pub struct APack {
	pub processed: bool,
	pub loaded: bool,
	pub compressed_bytes: Vec<u8>,
	pub vfs: HashMap<String, Vec<u8>>,
}

struct DebugVfs<'a>(&'a HashMap<String, Vec<u8>>);
impl<'a> Debug for DebugVfs<'a> {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		f.debug_map()
			.entries(self.0.iter().map(|(key, value)| {
				let key = key.clone();
				let value = format!("len({})", value.len());

				(key, value)
			}))
			.finish()
	}
}

impl Debug for APack {
	fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
		f.debug_struct("APack")
			.field("processed", &self.processed)
			.field(
				"compressed_bytes",
				&format!("len({})", self.compressed_bytes.len()),
			)
			.field("vfs", &DebugVfs(&self.vfs))
			.finish()
	}
}

pub struct ClonableAPack(pub APack);
impl Deref for ClonableAPack {
	type Target = APack;

	fn deref(&self) -> &Self::Target {
		&self.0
	}
}
impl DerefMut for ClonableAPack {
	fn deref_mut(&mut self) -> &mut Self::Target {
		&mut self.0
	}
}
impl Clone for ClonableAPack {
	fn clone(&self) -> Self {
		ClonableAPack(APack {
			loaded: self.loaded,
			processed: self.processed,
			compressed_bytes: self.compressed_bytes.clone(),
			vfs: self.vfs.clone(),
		})
	}
}
impl Into<APack> for ClonableAPack {
	fn into(self) -> APack {
		self.0
	}
}

impl APack {
	pub fn new(bytes: Vec<u8>) -> APack {
		APack {
			loaded: false,
			processed: false,
			compressed_bytes: bytes,
			vfs: Default::default(),
		}
	}

	/// Cloning an APack can be _very_ expensive, depending on the
	/// contents. To clone an APack, you've got to jump through some hoops
	pub fn into_clone(self) -> ClonableAPack {
		ClonableAPack(self)
	}

	pub fn clone_ref(apack: &APack) -> APack {
		APack {
			loaded: apack.loaded,
			processed: apack.processed,
			compressed_bytes: apack.compressed_bytes.clone(),
			vfs: apack.vfs.clone(),
		}
	}

	/// An APack doesn't need to hold onto the raw bytes after it has been processed.
	/// This method clears the raw buffer
	pub fn drop_raw_bytes(&mut self) {
		self.compressed_bytes = Vec::new();
	}

	pub fn set(&mut self, key: impl ToString, value: Vec<u8>) {
		self.vfs.insert(key.to_string(), value);
	}
}

impl Index<&String> for APack {
	type Output = Vec<u8>;

	fn index(&self, index: &String) -> &Self::Output {
		&self.vfs[index]
	}
}
impl IndexMut<&String> for APack {
	fn index_mut(&mut self, index: &String) -> &mut Self::Output {
		self.vfs
			.get_mut(index)
			.expect(&*format!("Missing key {}", index))
	}
}

impl AssetLoader for APackLoader {
	fn load<'a>(
		&'a self,
		bytes: &'a [u8],
		load_context: &'a mut LoadContext,
	) -> BoxedFuture<'a, anyhow::Result<(), anyhow::Error>> {
		Box::pin(async move {
			load_context.set_default_asset(LoadedAsset::new(APack::new(bytes.to_vec())));
			Ok(())
		})
	}

	fn extensions(&self) -> &[&str] {
		&["apack"]
	}
}

type TaskOutput = (APack, Handle<APack>);
#[derive(Component)]
pub struct ProcessAPack(Task<TaskOutput>);

fn create_task_body(
	mut pack: APack,
	handle_ref: Handle<APack>,
) -> impl futures_lite::Future<Output = TaskOutput> {
	async move {
		let mut pack = pack;

		let raw_bytes = std::mem::take(&mut pack.compressed_bytes);
		let decoder = XzDecoder::new(raw_bytes.as_slice());
		let mut tarboy = tar::Archive::new(decoder);

		if let Ok(entries) = tarboy.entries() {
			for entry in entries.flatten() {
				let path = &entry
					.path()
					.ok()
					.and_then(|v| v.to_str().map(String::from))
					.expect("Malformed path");

				let file = entry.bytes().flatten().collect::<Vec<u8>>();
				pack.set(path, file);
			}
		}

		pack.processed = true;
		pack.drop_raw_bytes();
		(pack, handle_ref)
	}
}

pub fn process_loaded_apacks(
	mut commands: Commands,
	mut events: EventReader<AssetEvent<APack>>,
	packs: Res<Assets<APack>>,
) {
	if !events.is_empty() {
		let pool = AsyncComputeTaskPool::get();

		for event in events.iter() {
			match event {
				AssetEvent::Created { handle } => {
					if let Some(pack) = packs.get(handle) {
						let copy = APack::clone_ref(pack);
						let handle_ref = handle.clone_weak();
						let task = pool.spawn(create_task_body(copy, handle_ref));

						commands.spawn(ProcessAPack(task));
					} else {
						log::warn!("Received a 'created' event for an apack that is not part of the Assets<APack> resource");
					}
				}
				AssetEvent::Modified { handle } => {
					if let Some(pack) = packs.get(handle) {
						if !pack.processed {
							log::debug!(
								"Handling hot-reload of APack by spawning new processing task"
							);

							let copy = APack::clone_ref(pack);
							let handle_ref = handle.clone_weak();
							let task = pool.spawn(create_task_body(copy, handle_ref));

							commands.spawn(ProcessAPack(task));
						} else {
							log::debug!("APack was changed, but will not notify because it has already been processed")
						}
					} else {
						log::warn!("Received a 'modified' event for an apack that is not part of the Assets<APack> resource");
					}
				}
				AssetEvent::Removed { .. } => {
					log::debug!("Removed an APack. Handles already extracted from this APack may not be removed");
				}
			}
		}
	}
}

pub struct APackProcessingComplete(pub Handle<APack>);

pub fn poll_apack_loading_tasks(
	mut commands: Commands,
	mut query: Query<(Entity, &mut ProcessAPack)>,
	mut packs: ResMut<Assets<APack>>,
	mut events: EventWriter<APackProcessingComplete>,
) {
	for (entity, mut task) in &mut query {
		if let Some((pack, handle)) = future::block_on(future::poll_once(&mut task.0)) {
			if let Some(existing) = packs.get_mut(&handle) {
				*existing = pack;
			}
			events.send(APackProcessingComplete(handle.clone_weak()));
			commands.entity(entity).despawn();
		}
	}
}
